import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_B3_TRADING_HOURS_B3
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog


class TradingHoursB3(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None,
        list_slugs:Optional[List[str]]=None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_B3_TRADING_HOURS_B3,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token,
        self.list_slugs = list_slugs

    def td_th_parser(self, resp_req:Response, list_th:List[Any]) \
        -> Tuple[List[Any], int, Optional[int]]:
        list_headers = list_th.copy()
        int_init_td = 0
        int_end_td = None
        if StrHandler().match_string_like(resp_req.url, '*#source=stocks*') == True:
            list_headers = [
                list_th[0],
                list_th[1] + ' Início',
                list_th[1] + ' Fim',
                list_th[2] + ' Início',
                list_th[2] + ' Fim',
                list_th[3] + ' Início',
                list_th[3] + ' Fim',
                list_th[4] + ' Início',
                list_th[4] + ' Fim',
                list_th[5] + ' ' + list_th[6] + ' Início',
                list_th[5] + ' ' + list_th[6] + ' Fim',
                list_th[5] + ' ' + list_th[7] + ' Início',
                list_th[5] + ' ' + list_th[7] + ' Fim',
                list_th[5] + ' ' + list_th[8] + ' Fechamento Início',
                list_th[5] + ' ' + list_th[8] + ' Fechamento Fim',
            ]
            int_init_td = 0
            int_end_td = 195
        elif StrHandler().match_string_like(resp_req.url, '*#source=stock_options*') == True:
            list_headers = [
                list_th[9],
                list_th[13] + ' Antes do Vencimento - Início',
                list_th[13] + ' Antes do Vencimento - Fim',
                list_th[13] + ' No Vencimento - Início',
                list_th[13] + ' No Vencimento - Fim',
                list_th[15] + ' No Vencimento',
                list_th[16] + ' No Vencimento - Início',
                list_th[16] + ' No Vencimento - Fim',
                list_th[17] + ' No Vencimento - Início',
            ]
            int_init_td = 195
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=pmi_future*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing',
                list_th[7] + ' - Opening',
                list_th[7] + ' - Closing',
            ]
            int_init_td = 0
            int_end_td = 52
        elif StrHandler().match_string_like(resp_req.url, '*#source=stock_index_futures*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing',
                list_th[7] + ' - Opening',
                list_th[7] + ' - Closing',
            ]
            int_init_td = 52
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=interest_rates*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing',
                list_th[7] + ' - Opening',
                list_th[7] + ' - Closing',
            ]
            int_init_td = 0
            int_end_td = 130
        elif StrHandler().match_string_like(resp_req.url, '*#source=usd_interest_rates*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing',
                list_th[7] + ' - Opening',
                list_th[7] + ' - Closing',
            ]
            int_init_td = 130
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=commodities*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing'
            ]
            int_init_td = 0
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=crypto*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing'
            ]
            int_init_td = 0
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=foreign_exchange_and_dollar_spot*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2] + ' - Opening',
                list_th[2] + ' - Closing',
                list_th[3] + ' - Opening',
                list_th[3] + ' - Closing',
                list_th[4] + ' - Opening',
                list_th[5] + ' EOD - Opening',
                list_th[5] + ' EOD - Closing',
                list_th[6] + ' - Opening',
                list_th[6] + ' - Closing',
                list_th[7] + ' - Opening',
                list_th[7] + ' - Closing',
            ]
            int_init_td = 0
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=otc*') == True:
            list_headers = [
                list_th[0],
                list_th[1] + ' - Início',
                list_th[1] + ' - Fim',
                list_th[2] + ' - Início',
                list_th[2] + ' - Fim',
                list_th[3] + ' - Início',
                list_th[3] + ' - Fim',
                list_th[10] + ' - Antes do Vencimento - Exercício de Posição Titular - Início',
                list_th[10] + ' - Antes do Vencimento - Exercício de Posição Titular - Fim',
                list_th[10] + ' - No Vencimento - Encerramento de Posição - Início',
                list_th[10] + ' - No Vencimento - Encerramento de Posição - Fim',
                list_th[10] + ' - No Vencimento - Exercício de Posição Titular - Início',
                list_th[10] + ' - No Vencimento - Exercício de Posição Titular - Fim',
                list_th[4] + ' - Opening',
                list_th[4] + ' - Closing',
                list_th[6] + ' EOD - Opening',
                list_th[6] + ' EOD - Closing',
                list_th[5] + ' - Opening',
                list_th[5] + ' - Closing',
            ]
            int_init_td = 0
            int_end_td = None
        elif StrHandler().match_string_like(resp_req.url, '*#source=opf_before_exc_date*') == True:
            list_headers = [
                list_th[0],
                list_th[1],
                list_th[2],
                list_th[3]
            ]
            int_init_td = 0
            int_end_td = 32
        elif StrHandler().match_string_like(resp_req.url, '*#source=opf_after_exc_date*') == True:
            list_headers = [
                list_th[4],
                list_th[5],
                list_th[6],
                list_th[7] + ' Without Exercise Risk',
                list_th[7] + ' Holder'
            ]
            int_init_td = 32
            int_end_td = None
        else:
            if self.logger is not None:
                CreateLog().warning(
                    self.logger,
                    'No source found in url, for HTML webscraping, please revisit the code'
                    + f' if it is an unexpected behaviour - URL: {resp_req.url}'
                )
            else:
                print('No source found in url, for HTML webscraping, please revisit the code'
                    + f' if it is an unexpected behaviour - URL: {resp_req.url}')
        return list_headers, int_init_td, int_end_td

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        bl_debug = True if StrHandler().match_string_like(
            resp_req.url, '*&bl_debug=True*') == True else False
        root = HtmlHandler().lxml_parser(resp_req)
        # export html tree to data folder, if is user's will
        if bl_debug == True:
            path_project = DirFilesManagement().find_project_root(marker='pyproject.toml')
            HtmlHandler().html_tree(root, file_path=rf'{path_project}/data/test.html')
        list_th = [
            x.text.strip() for x in HtmlHandler().lxml_xpath(
                root, YAML_B3_TRADING_HOURS_B3['stocks']['xpaths']['list_th']
            )
        ]
        list_td = [
            '' if x.text is None else x.text.replace('\xa0', '').strip()
            for x in HtmlHandler().lxml_xpath(
                root, YAML_B3_TRADING_HOURS_B3['stocks']['xpaths']['list_td']
            )
        ]
        if bl_debug == True:
            print(f'LIST TH RAW: {list_th}')
            print(f'LIST TD RAW: {list_td}')
        # deal with data/headers specificity for the project
        list_headers, int_init_td, int_end_td = self.td_th_parser(resp_req, list_th)
        if bl_debug == True:
            print(f'LIST TH TRT: {list_headers}')
            print(f'LEN LIST HEADERS: {len(list_headers)}')
            print(f'LIST TD TRT: {list_td[int_init_td:int_end_td]}')
            print(f'LEN LIST TD: {len(list_td[int_init_td:int_end_td])}')
        list_ser = HandlingDicts().pair_headers_with_data(
            list_headers,
            list_td[int_init_td:int_end_td]
        )
        return pd.DataFrame(list_ser)



### DIMENSIONALITY REDUCTION OF IVS

# pypi.org libs
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestClassifier
from mlxtend.feature_selection import SequentialFeatureSelector, ExhaustiveFeatureSelector
from sklearn.decomposition import PCA
# local libs
from stpstone.transformations.cleaner.eda import ExploratoryDataAnalysis


class DimensionalityReduction:

    def backward_elimination(self, array_x, array_y, str_estimator='linear_regression', int_cv=5,
                             int_verbose=0):
        """
        REFERENCES:
            https://rasbt.github.io/mlxtend/api_subpackages/mlxtend.feature_selection/#sequentialfeatureselector,
            https://rasbt.github.io/mlxtend/user_guide/feature_selection/SequentialFeatureSelector/
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # defining the estimator
        if str_estimator == 'linear_regression':
            class_estimator = LinearRegression()
            scoring = 'r2'
        elif str_estimator == 'rf_classifier':
            class_estimator = RandomForestClassifier(n_jobs=-1)
            scoring = 'accuracy'
        else:
            raise Exception('Estimator {} not defined, please revisit this parameter'.format(
                str_estimator
            ))
        # running the backward feature selection
        feature_selection = SequentialFeatureSelector(
            class_estimator,
            k_features=(1, array_x.shape[1]),
            forward=False,
            floating=False,
            verbose=int_verbose,
            scoring=scoring,
            cv=int_cv
        ).fit(array_x, array_y)
        # returning the most relevant columns
        return {
            'dr_feature_names': feature_selection.k_feature_names_,
            'dr_feature_indexes': feature_selection.k_feature_idx_,
            'dr_score': feature_selection.k_score_
        }

    def forward_elimination(self, array_x, array_y, str_estimator='linear_regression', int_cv=5,
                            int_verbose=0):
        """
        REFERENCES:
            https://rasbt.github.io/mlxtend/api_subpackages/mlxtend.feature_selection/#sequentialfeatureselector,
            https://rasbt.github.io/mlxtend/user_guide/feature_selection/SequentialFeatureSelector/
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # defining the estimator
        if str_estimator == 'linear_regression':
            class_estimator = LinearRegression()
            scoring = 'r2'
        elif str_estimator == 'rf_classifier':
            class_estimator = RandomForestClassifier()
            scoring = 'accuracy'
        else:
            raise Exception('Estimator {} not defined, please revisit this parameter'.format(
                str_estimator
            ))
        # running the backward feature selection
        feature_selection = SequentialFeatureSelector(
            class_estimator,
            k_features=(1, array_x.shape[1]),
            forward=True,
            floating=False,
            verbose=int_verbose,
            scoring=scoring,
            cv=int_cv
        ).fit(array_x, array_y)
        # returning the most relevant columns
        return {
            'dr_feature_names': feature_selection.k_feature_names_,
            'dr_feature_indexes': feature_selection.k_feature_idx_,
            'dr_score': feature_selection.k_score_
        }

    def exhaustive_feature_selection(self, array_x, array_y, str_estimator='linear_regression',
                                     max_features=None, int_cv=5):
        """
        REFERENCES:
            https://rasbt.github.io/mlxtend/api_subpackages/mlxtend.feature_selection/#exhaustivefeatureselector
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # maximum features
        if max_features is None:
            max_features=array_x.shape[1]
        # defining the estimator
        if str_estimator == 'linear_regression':
            class_estimator = LinearRegression()
            scoring = 'r2'
        elif str_estimator == 'rf_classifier':
            class_estimator = RandomForestClassifier()
            scoring = 'accuracy'
        else:
            raise Exception('Estimator {} not defined, please revisit this parameter'.format(
                str_estimator
            ))
        # running the backward feature selection
        feature_selection = ExhaustiveFeatureSelector(
            class_estimator,
            min_features=1,
            max_features=max_features,
            scoring=scoring,
            cv=int_cv,
            n_jobs=-1
        ).fit(array_x, array_y)
        # returning the most relevant columns
        return {
            'dr_feature_names': feature_selection.best_feature_names_ ,
            'dr_feature_indexes': feature_selection.best_idx_,
            'dr_score': feature_selection.best_score_
        }

    def iv_woe_iter(self, df_binned, target_col, class_col, name_missing_data='Missing',
                    name_bins='_bins', col_min='min', col_max='max',
                    col_non_event_count='non_event_count', col_sample_class='sample_class',
                    col_event_count='event_count', col_min_value='min_value',
                    col_max_value='max_value', col_sample_count='sample_count',
                    col_event_rate='event_rate', col_non_event_rate='non_event_rate',
                    col_feature='feature', col_sample_class_label='sample_class_label',
                    name_category='category', col_pct_non_event='pct_non_event',
                    col_pct_event='pct_event', col_first='first', col_count='count',
                    col_sum='sum', col_mean='mean', col_woe='woe', col_iv='iv'):
        """
        REFERENCES: https://github.com/pankajkalania/IV-WOE/blob/main/iv_woe_code.py,
            https://gaurabdas.medium.com/weight-of-evidence-and-information-value-in-python-from-scratch-a8953d40e34#:~:text=Information%20Value%20gives%20us%20the,as%20good%20and%20bad%20customers.
        DOCSTRING: INFORMATION VALUE (IV) AND WEIGHT OF EVIDENCE (WOE) USED TO UNDERSTAND
            THE RELATION BETWEEN DEPENDENT AND INDEPENDENT VARIABLES - PROVIDE THE STRENGTH OF
            A PREDICTOR TO MEASURE BINARY OUTCOMES
        INPUTS:
        OUTPUTS:
        """
        # replacing bins column names
        if name_bins in class_col:
            #   handle missing data
            df_binned[class_col] = df_binned[class_col].cat.add_categories([
                                                                           name_missing_data])
            df_binned[class_col] = df_binned[class_col].fillna(
                name_missing_data)
            #   dealing with _bins part of columns names
            df_tmp_groupby = df_binned.groupby(class_col).agg(
                {
                    class_col.replace(name_bins, ''): [col_min, col_max],
                    target_col: [col_count, col_sum, col_mean]
                }
            ).reset_index()
        else:
            #   handle missing data
            df_binned[class_col] = df_binned[class_col].fillna(
                name_missing_data)
            #   dealing with _bins part of columns names
            df_tmp_groupby = df_binned.groupby(class_col).agg(
                {
                    class_col: [col_first, col_first],
                    target_col: [col_count, col_sum, col_mean]
                }
            ).reset_index()
        # renaming colums of interest
        df_tmp_groupby.columns = [col_sample_class, col_min_value, col_max_value, col_sample_count,
                                  col_event_count, col_event_rate]
        # non event countage
        df_tmp_groupby[col_non_event_count] = df_tmp_groupby[col_sample_count] \
            - df_tmp_groupby[col_event_count]
        # non event rate
        df_tmp_groupby[col_non_event_rate] = 1 - df_tmp_groupby[col_event_rate]
        # limiting columns of interest
        df_tmp_groupby = df_tmp_groupby[[col_sample_class, col_min_value, col_max_value,
                                         col_sample_count,
                                         col_non_event_count, col_non_event_rate, col_event_count,
                                         col_event_rate]]
        # dealing with missing data
        if name_bins not in class_col and name_missing_data in df_tmp_groupby[col_min_value]:
            df_tmp_groupby[col_min_value] = df_tmp_groupby[col_min_value].replace({
                name_missing_data: np.nan})
            df_tmp_groupby[col_max_value] = df_tmp_groupby[col_max_value].replace({
                name_missing_data: np.nan
            })
        # creating feature column
        df_tmp_groupby[col_feature] = class_col
        # sampling labels
        if name_bins in class_col:
            df_tmp_groupby[col_sample_class_label] = df_tmp_groupby[col_sample_class].replace({
                name_missing_data: np.nan
            }).astype(name_category).cat.codes.replace({-1: np.nan})
        else:
            df_tmp_groupby[col_sample_class_label] = np.nan
        # limiting columns of interest
        df_tmp_groupby = df_tmp_groupby[[col_feature, col_sample_class, col_sample_class_label,
                                         col_sample_count, col_min_value, col_max_value,
                                         col_non_event_count,
                                         col_non_event_rate, col_event_count, col_event_rate]]
        # get percentage of wanted and unwanted evnts
        df_tmp_groupby[col_pct_non_event] = df_tmp_groupby[col_non_event_count] \
            / df_tmp_groupby[col_non_event_count].sum()
        df_tmp_groupby[col_pct_event] = df_tmp_groupby[col_event_count] \
            / df_tmp_groupby[col_event_count].sum()
        # calculating weight of evidence and information value
        df_tmp_groupby[col_woe] = np.log(df_tmp_groupby[col_pct_non_event]
                                         / df_tmp_groupby[col_pct_event])
        df_tmp_groupby[col_iv] = (df_tmp_groupby[col_pct_non_event]
                                  - df_tmp_groupby[col_pct_event]) * df_tmp_groupby[col_woe]
        # dealing with infinite numbers
        df_tmp_groupby[col_woe] = df_tmp_groupby[col_woe].replace(
            [np.inf, -np.inf], 0)
        df_tmp_groupby[col_iv] = df_tmp_groupby[col_iv].replace(
            [np.inf, -np.inf], 0)
        # returning datframe of interest
        return df_tmp_groupby

    def var_iter(self, df, target_col, max_bins,
                 min_bins_monotonic=2, col_feature='feature', col_remarks='remarks',
                 name_categorical='categorical', orient_str='records'):
        """
        REFERENCES: https://github.com/pankajkalania/IV-WOE/blob/main/iv_woe_code.py,
            https://gaurabdas.medium.com/weight-of-evidence-and-information-value-in-python-from-scratch-a8953d40e34#:~:text=Information%20Value%20gives%20us%20the,as%20good%20and%20bad%20customers.
        DOCSTRING: ITERATE OVER ALL FEATURES, CALCULATE WOE AND IV FOR THE CLASSES, AND
            APPEND TO ONE DATAFRAME WOE_IV OF EXPORTATION
        INPUTS:
        OUTPUTS:
        """
        # setting variables
        list_ser_woe_iv = list()
        list_remarks = list()
        # iterating over the columns within the dataframe
        for c_i in df.columns:
            if c_i not in [target_col]:
                #   check if binning is required, if so, then prepare bins and calculate woe and
                #       iv - logic: binning is done only when feature is continuous and non-binary;
                #       note: make sure dtype of continuous columns in dataframe is not object
                if np.issubdtype(df[c_i], np.number) and df[c_i].nunique() > min_bins_monotonic:
                    class_col, remarks, df_binned = self.prepare_bins(
                        df[[c_i, target_col]].copy(), c_i, target_col, max_bins)
                    #   calculating iv and woe
                    df_agg_data = self.iv_woe_iter(
                        df_binned.copy(), target_col, class_col)
                    list_remarks.append(
                        {col_feature: c_i, col_remarks: remarks})
                else:
                    df_agg_data = self.iv_woe_iter(
                        df[[c_i, target_col]].copy(), target_col, c_i)
                    list_remarks.append(
                        {col_feature: c_i, col_remarks: name_categorical})
                #   appending to list of dataframes
                list_ser_woe_iv.extend(
                    df_agg_data.to_dict(orient=orient_str))
        # returning dataframes of woe, iv and remarks
        return pd.DataFrame(list_ser_woe_iv), pd.DataFrame(list_remarks)

    def get_iv_woe(self, df, target_col, max_bins, name_missing_data='Missing', name_bins='_bins',
                   col_non_event_count='non_event_count', col_sample_class='sample_class',
                   col_event_count='event_count', col_min_value='min_value',
                   col_max_value='max_value', col_sample_count='sample_count',
                   col_non_event_rate='non_event_rate', col_feature='feature',
                   col_sample_class_label='sample_class_label', col_pct_non_event='pct_non_event',
                   col_pct_event='pct_event', col_count='count', col_woe='woe', col_sum='sum',
                   col_iv='iv', col_remarks='remarks', col_number_classes='number_of_classes',
                   col_feature_null_pct='feature_null_percent', col_iv_sum='iv_sum'):
        """
        REFERENCES: https://github.com/pankajkalania/IV-WOE/blob/main/iv_woe_code.py,
            https://gaurabdas.medium.com/weight-of-evidence-and-information-value-in-python-from-scratch-a8953d40e34#:~:text=Information%20Value%20gives%20us%20the,as%20good%20and%20bad%20customers.
        DOCSTRING: AFTER GETTING WOE AND IV FOR ALL CLASSES OF FEATURES, CALCULATE THE AGGREGATED
            IV VALUES FOR FEATURES
        INPUTS:
        OUTPUTS:
        """
        # calculating woe, iv and getting the remarks for each bin
        df_woe_iv, df_binning_remarks = self.var_iter(df, target_col, max_bins)
        # limiting the columns of interest
        df_woe_iv[col_feature] = df_woe_iv[col_feature].replace(
            name_bins, '', regex=True)
        df_woe_iv = df_woe_iv[[col_feature, col_sample_class, col_sample_class_label, col_sample_count,
                               col_min_value, col_max_value, col_non_event_count, col_non_event_rate,
                               col_event_count, col_pct_non_event, col_pct_event, col_woe, col_iv]]
        # getting the accumulated iv for each feature
        df_iv = df_woe_iv.groupby(col_feature)[[col_iv]].agg(
            [col_sum, col_count]).reset_index()
        # changing column names
        df_iv.columns = [col_feature, col_iv, col_number_classes]
        # handling null values
        df_null_percent = pd.DataFrame(df.isnull().mean()).reset_index()
        df_null_percent.columns = [col_feature, col_feature_null_pct]
        # left-join iv with null values
        df_iv = df_iv.merge(df_null_percent, on=col_feature, how='left')
        # left-join iv with binning remarks
        df_iv = df_iv.merge(df_binning_remarks, on=col_feature, how='left')
        # left-join woe_iv dataframe with aggregated iv
        df_woe_iv = df_woe_iv.merge(df_iv[[col_feature, col_iv, col_remarks]].rename(columns={
            col_iv: col_iv_sum}), on=col_feature, how='left')
        # remove duplicates
        df_iv.drop_duplicates(inplace=True)
        # sort values
        df_iv.sort_values([col_iv], ascending=[False], inplace=True)
        # return iv and woe_iv dataframe
        return df_iv, df_woe_iv.replace({name_missing_data: np.nan})

    def iv_label_predictive_power(self, df_iv, col_iv='iv', col_predictive_power='predictive_power'):
        """
        REFERENCES: https://github.com/pankajkalania/IV-WOE/blob/main/iv_woe_code.py,
            https://gaurabdas.medium.com/weight-of-evidence-and-information-value-in-python-from-scratch-a8953d40e34#:~:text=Information%20Value%20gives%20us%20the,as%20good%20and%20bad%20customers.
        DOCSTRING: LABEL PREDICTIVE POWER OF EACH FEATURE
        INPUTS: DATAFRAME INFORMATION VALUE(IV)
        OUTPUTS: DATAFRAME
        """
        # sort values
        df_iv.sort_values([col_iv], ascending=[False], inplace=True)
        # set label strength
        for index, row in df_iv.iterrows():
            if row[col_iv] < 0.02:
                df_iv.loc[index, col_predictive_power] = 'not useful'
            elif row[col_iv] < 0.1:
                df_iv.loc[index, col_predictive_power] = 'weak predictor'
            elif row[col_iv] < 0.3:
                df_iv.loc[index, col_predictive_power] = 'medium predictor'
            elif row[col_iv] < 0.5:
                df_iv.loc[index, col_predictive_power] = 'strong predictor'
            else:
                df_iv.loc[index, col_predictive_power] = 'suspicious'
        # return dataframe iv
        return df_iv

    def pca(self, array_data):
        """
        REFERENCES: https://leandrocruvinel.medium.com/pca-na-mão-e-no-python-d559e9c8f053
        DOCSTRING: PRINCIPAL COMPONENTS ANALYSIS
        INPUTS: DEPENDENT AND INDEPENDENT VARIABLES
        OUTPUTS:
        """
        # pca fiiting model
        model_fitted = PCA(n_components=len(array_data)).fit(array_data)
        # returning model infos
        return {
            'eigenvalues': model_fitted.explained_variance_,
            'eigenvectors': model_fitted.components_,
            'explained_variance_ratio': model_fitted.explained_variance_ratio_,
            'components': model_fitted.components_
        }



### USA WORKING DAYS ###

from datetime import datetime, timedelta
from workalendar.usa import UnitedStates


class WorkCalendar:

    def __init__(self) -> None:
        self.calendar = UnitedStates()

    def add_working_days(self, dt_bgn:datetime, int_num_days:int) -> datetime:
        current_date = dt_bgn
        working_days_added = 0
        while working_days_added < int_num_days:
            current_date += timedelta(days=1)
            if current_date.weekday() < 5 and not self.calendar.is_holiday(current_date):
                working_days_added += 1
        return current_date

    def is_holiday(self, date:datetime) -> bool:
        return self.calendar.is_holiday(date)

    def is_weekend(self, date:datetime) -> bool:
        return date.weekday() >= 5

    def diff_working_days(self, dt_bgn:datetime, dt_end:datetime) -> int:
        current_date = dt_bgn
        working_days = 0
        while current_date <= dt_end:
            if current_date.weekday() < 5 and not self.calendar.is_holiday(current_date):
                working_days += 1
            current_date += timedelta(days=1)
        return working_days


if __name__ == '__main__':

    cls_usabzdays = WorkCalendar()
    # current date
    dt_bgn = datetime(2025, 1, 1)
    # add 10 working days to the start date
    new_date = cls_usabzdays.add_working_days(dt_bgn, 10)
    print(f"Date after adding 10 working days: {new_date}")
    # check if a specific date is a holiday
    check_date = datetime(2025, 1, 20)  # Martin Luther King Jr. Day (USA)
    print(f"Is {check_date} a holiday? {cls_usabzdays.is_holiday(check_date)}")
    # check if a specific date is a weekend
    weekdt_sup = datetime(2025, 1, 11)
    print(f"Is {weekdt_sup} a weekend? {cls_usabzdays.is_weekend(weekdt_sup)}")



import os
from stpstone.utils.parsers.yaml import reading_yaml


root_path = os.path.dirname(os.path.realpath(__file__))


# * generic
YAML_GEN = reading_yaml(os.path.join(root_path, "generic.yaml"))

# * bylaws
YAML_INVESTMENT_FUNDS_BYLAWS = reading_yaml(
    os.path.join(root_path, "countries/br/bylaws/investment_funds.yaml")
)

# * exchange
#   BR
YAML_B3_UP2DATA_REGISTRIES = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/up2data_registries.yaml"))
YAML_B3_UP2DATA_VOLUMES_TRD = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/up2data_volumes_trd.yaml"))
YAML_B3_SEARCH_BY_TRADING_SESSION = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/search_by_trading_session.yaml"))
YAML_B3_BVMF_BOV = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/bvmf_bov.yaml"))
YAML_B3_OPTIONS_CALENDAR = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/options_calendar.yaml"))
YAML_B3_WARRANTY = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/warranty.yaml"))
YAML_B3_TRADING_HOURS_B3 = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/trading_hours.yaml"))
YAML_B3_INDEXES_THEOR_PORTF = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/indexes_theor_portf.yaml"))
YAML_B3_HISTORICAL_SIGMA = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/historical_sigma.yaml"))
YAML_B3_CONSOLIDATED_TRDS = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/consolidated_trades.yaml"))
YAML_B3_CONSOLIDATED_TRDS_AFTER_MKT = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/consolidated_trades_after_mkt.yaml"))
YAML_B3_FUTURES_CLOSING_ADJ = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/futures_closing_adj.yaml"))
YAML_ANBIMA_DATA_INDEXES = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/indexes_anbima_data.yaml"))
YAML_ANBIMA_INDEXES = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/indexes_anbima.yaml"))
YAML_ANBIMA_550_LISTING = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/anbima_550_listing.yaml"))
YAML_BMF_INTEREST_RATES = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/bmf_interest_rates.yaml"))
YAML_ANBIMA_INFOS = reading_yaml(
    os.path.join(root_path, "countries/br/exchange/anbima_site_infos.yaml"))
#   US
YAML_US_ALPHAVANTAGE = reading_yaml(
    os.path.join(root_path, "countries/us/exchange/alphavantage.yaml"))
YAML_US_TIINGO = reading_yaml(
    os.path.join(root_path, "countries/us/exchange/tiingo.yaml"))
#   WW
YAML_WW_CRYPTO_COINMARKET = reading_yaml(
    os.path.join(root_path, "countries/ww/exchange/crypto/coinmarket.yaml"))
YAML_WW_CRYPTO_COINPAPRIKA = reading_yaml(
    os.path.join(root_path, "countries/ww/exchange/crypto/coinpaprika.yaml"))
YAML_WW_CRYPTO_COINCAP = reading_yaml(
    os.path.join(root_path, "countries/ww/exchange/crypto/coincap.yaml"))
YAML_WW_FMP = reading_yaml(
    os.path.join(root_path, "countries/ww/exchange/markets/fmp.yaml"))
YAML_WW_INVESTINGCOM = reading_yaml(
    os.path.join(root_path, "countries/ww/exchange/markets/investingcom.yaml"))

# * macroeconomics
# BR
YAML_BR_PTAX_BCB = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/ptax_bcb.yaml"))
YAML_ANBIMA_FORECASTS = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/anbima_pmi_forecasts.yaml"))
YAML_YAHII_RATES = reading_yaml(os.path.join(root_path, "countries/br/macroeconomics/yahii_rates.yaml"))
YAML_YAHII_OHTERS = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/yahii_others.yaml"))
YAML_SIDRA_IBGE = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/sidra_ibge.yaml"))
YAML_SGS_BCB = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/sgs_bcb.yaml"))
YAML_B3_FINANCIAL_INDICATORS = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/b3_financial_indicators.yaml"))
YAML_OLINDA_BCB = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/olinda_bcb.yaml"))
YAML_INVESTINGCOM_BR = reading_yaml(
    os.path.join(root_path, "countries/br/macroeconomics/investingcom_br.yaml"))
# US
YAML_FRED_US = reading_yaml(
    os.path.join(root_path, "countries/us/macroeconomics/fred.yaml"))
# WW
YAML_WW_GR = reading_yaml(
    os.path.join(root_path, "countries/ww/macroeconomics/global_rates.yaml"))
YAML_WW_TRADING_ECON = reading_yaml(
    os.path.join(root_path, "countries/ww/macroeconomics/trading_economics.yaml"))
YAML_WW_WORLD_GOV_BONDS = reading_yaml(
    os.path.join(root_path, "countries/ww/macroeconomics/world_gov_bonds.yaml"))
YAML_WW_RATINGS_AGENCIES = reading_yaml(
    os.path.join(root_path, "countries/ww/macroeconomics/ratings_agencies.yaml"))

# * otc
# BR
YAML_DEBENTURES = reading_yaml(
    os.path.join(root_path, "countries/br/otc/debentures.yaml"))

# * registries
# BR
YAML_BR_CVM_REGISTRIES = reading_yaml(
    os.path.join(root_path, "countries/br/registries/cvm.yaml"))
YAML_BR_BCB_BRAZILLIAN_BANKS_REGISTRY = reading_yaml(
    os.path.join(root_path, "countries/br/registries/brazillian_banks.yaml"))
YAML_B3_TRD_SEC = reading_yaml(
    os.path.join(root_path, "countries/br/registries/b3_trd_sec.yaml"))
YAML_ANBIMA_DATA_FUNDS = reading_yaml(
    os.path.join(root_path, "countries/br/registries/anbima_data_funds.yaml"))
YAML_ANBIMA_DATA_DEBENTURES = reading_yaml(
    os.path.join(root_path, "countries/br/registries/anbima_data_debentures.yaml"))
YAML_ANBIMA_DATA_API = reading_yaml(
    os.path.join(root_path, "countries/br/registries/anbima_data_api.yaml"))
YAML_MAIS_RETORNO_FUNDS = reading_yaml(
    os.path.join(root_path, "countries/br/registries/mais_retorno_instruments.yaml"))
# US
YAML_US_SLICKCHARTS_INDEXES_COMPONENTS = reading_yaml(
    os.path.join(root_path, "countries/us/registries/slickcharts_indexes_components.yaml"))
YAML_US_ETFDB_VETTAFI = reading_yaml(
    os.path.join(root_path, "countries/us/registries/etfdb_vettafi.yaml"))
# WW
YAML_WW_RATINGS_CORP_S_AND_P = reading_yaml(
    os.path.join(root_path, "countries/ww/registries/ratings_corp_spglobal.yaml"))

# * taxation
# BR
YAML_IRSBR = reading_yaml(os.path.join(root_path, "countries/br/taxation/irsbr.yaml"))



import os
import base64
import hashlib
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding as sym_padding
from cryptography.hazmat.backends import default_backend
import bcrypt


class SecureCrypto:
    """
    A class providing secure cryptographic operations including:
    - bcrypt password hashing and verification
    - PBKDF2 key derivation
    - SHA-256 hashing
    - AES-256 encryption/decryption (CBC mode with random IV)
    - Secure random token generation
    """
    
    # AES block size in bytes
    AES_BLOCK_SIZE = 16
    # PBKDF2 iterations (recommended minimum is 600,000 as of 2023)
    PBKDF2_ITERATIONS = 600000
    
    def __init__(self, pbkdf2_iterations=None):
        """Initialize the crypto instance with optional custom parameters"""
        if pbkdf2_iterations is not None:
            self.PBKDF2_ITERATIONS = pbkdf2_iterations
    
    def generate_salt(self, size=16):
        """Generate a cryptographically secure random salt"""
        return os.urandom(size)
    
    def bcrypt_hash_password(self, password: str) -> str:
        """
        Hash a password using bcrypt with automatically generated salt.
        Returns a string containing the hashed password.
        """
        if not isinstance(password, str):
            raise TypeError("Password must be a string")
        
        # Convert to bytes and hash with bcrypt
        password_bytes = password.encode('utf-8')
        hashed = bcrypt.hashpw(password_bytes, bcrypt.gensalt())
        return hashed.decode('utf-8')
    
    def bcrypt_verify_password(self, password: str, hashed_password: str) -> bool:
        """
        Verify a password against a bcrypt hashed password.
        Returns True if the password matches, False otherwise.
        """
        try:
            password_bytes = password.encode('utf-8')
            hashed_bytes = hashed_password.encode('utf-8')
            return bcrypt.checkpw(password_bytes, hashed_bytes)
        except (ValueError, TypeError):
            return False
    
    def pbkdf2_derive_key(self, password: str, salt: bytes = None, iterations: int = None) -> tuple:
        """
        Derive a cryptographic key using PBKDF2-HMAC-SHA256.
        Returns a tuple of (derived_key_bytes, salt_used).
        """
        if salt is None:
            salt = self.generate_salt(16)
        if iterations is None:
            iterations = self.PBKDF2_ITERATIONS
        
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,  # 32 bytes = 256 bits
            salt=salt,
            iterations=iterations,
            backend=default_backend()
        )
        key = kdf.derive(password.encode('utf-8'))
        return (key, salt)
    
    def sha256_hash(self, data: str) -> str:
        """Generate SHA-256 hash of the input data"""
        if not isinstance(data, str):
            raise TypeError("Data must be a string")
        
        sha256 = hashlib.sha256()
        sha256.update(data.encode('utf-8'))
        return sha256.hexdigest()
    
    def aes_encrypt(self, plaintext: str, key: bytes) -> str:
        """
        Encrypt plaintext using AES-256 in CBC mode with PKCS7 padding.
        Returns a base64 encoded string containing the IV and ciphertext.
        """
        if not isinstance(plaintext, str):
            raise TypeError("Plaintext must be a string")
        if len(key) != 32:
            raise ValueError("Key must be 32 bytes (256 bits) for AES-256")
        
        # Generate random IV
        iv = os.urandom(self.AES_BLOCK_SIZE)
        
        # Pad the plaintext
        padder = sym_padding.PKCS7(self.AES_BLOCK_SIZE * 8).padder()  # Fixed padding import
        padded_data = padder.update(plaintext.encode('utf-8')) + padder.finalize()
        
        # Encrypt
        cipher = Cipher(
            algorithms.AES(key),
            modes.CBC(iv),
            backend=default_backend()
        )
        encryptor = cipher.encryptor()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()
        
        # Combine IV and ciphertext and return as base64
        combined = iv + ciphertext
        return base64.b64encode(combined).decode('utf-8')
    
    def aes_decrypt(self, encrypted_data: str, key: bytes) -> str:
        """
        Decrypt data encrypted with aes_encrypt().
        Returns the decrypted plaintext string.
        """
        if len(key) != 32:
            raise ValueError("Key must be 32 bytes (256 bits) for AES-256")
        
        # Decode base64 and split IV from ciphertext
        combined = base64.b64decode(encrypted_data)
        iv = combined[:self.AES_BLOCK_SIZE]
        ciphertext = combined[self.AES_BLOCK_SIZE:]
        
        # Decrypt
        cipher = Cipher(
            algorithms.AES(key),
            modes.CBC(iv),
            backend=default_backend()
        )
        decryptor = cipher.decryptor()
        padded_plaintext = decryptor.update(ciphertext) + decryptor.finalize()
        
        # Unpad
        unpadder = sym_padding.PKCS7(self.AES_BLOCK_SIZE * 8).unpadder()  # Fixed padding import
        plaintext = unpadder.update(padded_plaintext) + unpadder.finalize()
        
        return plaintext.decode('utf-8')
    
    def generate_secure_token(self, length=32) -> str:
        """
        Generate a cryptographically secure random token.
        Returns a URL-safe base64 encoded string.
        """
        if length < 16:
            raise ValueError("Token length should be at least 16 bytes for security")
        
        random_bytes = os.urandom(length)
        return base64.urlsafe_b64encode(random_bytes).decode('utf-8').rstrip('=')
    
    def encrypt_with_password(self, plaintext: str, password: str) -> str:
        """
        High-level method to encrypt data with a password.
        Uses PBKDF2 to derive a key from the password, then AES-256 encryption.
        Returns a base64 encoded string containing salt, IV, and ciphertext.
        """
        # Derive key from password
        key, salt = self.pbkdf2_derive_key(password)
        
        # Encrypt the data
        encrypted = self.aes_encrypt(plaintext, key)
        
        # Combine salt and encrypted data
        combined = salt + base64.b64decode(encrypted)
        return base64.b64encode(combined).decode('utf-8')
    
    def decrypt_with_password(self, encrypted_data: str, password: str) -> str:
        """
        Decrypt data encrypted with encrypt_with_password().
        Returns the decrypted plaintext string.
        """
        # Decode base64 and split components
        combined = base64.b64decode(encrypted_data)
        salt = combined[:16]  # Salt size matches what pbkdf2_derive_key uses
        encrypted_part = combined[16:]
        
        # Derive key from password and salt
        key, _ = self.pbkdf2_derive_key(password, salt)
        
        # Decrypt the data
        return self.aes_decrypt(base64.b64encode(encrypted_part).decode('utf-8'), key)


import time
import pandas as pd
from pprint import pprint
from typing import Any, Dict, List, Optional, Union
from requests import request
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.parsers.json import JsonFiles


class ConnectionApi(metaclass=TypeChecker):
    """
    Metadata:
        - http://www.b3.com.br/data/files/2E/95/28/F1/EBD17610515A8076AC094EA8/GUIDE-TO-LINE-5.0-API.pdf,
        - https://line.bvmfnet.com.br/#/endpoints

    Notes: Please contact CAU manager account to request a service user
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        broker_code: str,
        category_code: str,
        hostname_api_line_b3: str = "https://api.line.bvmfnet.com.br",
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.broker_code = broker_code
        self.category_code = category_code
        self.hostname_api_line_b3 = hostname_api_line_b3
        self.token = self.access_token

    @property
    def auth_header(
        self,
        method: str = "GET",
        key_header: str = "header",
        int_max_retrieves: int = 1000,
        int_status_code_ok: int = 200,
        int_status_code_iteration: int = 400,
        bl_verify: bool = False,
        app: str = "/api/v1.0/token/authorization",
    ) -> str:
        # passing variables
        i = 0
        # requesting authorization authheader
        dict_headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        # looping while status code not a 2xx response
        while (int_status_code_iteration != int_status_code_ok) and (
            i <= int_max_retrieves
        ):
            try:
                resp_req = request(
                    method=method,
                    url=self.hostname_api_line_b3 + app,
                    headers=dict_headers,
                    verify=bl_verify,
                )
            except:
                continue
            int_status_code_iteration = resp_req.status_code
            i += 1
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # getting authheader
        return resp_req.json()[key_header]

    @property
    def access_token(
        self,
        method: str = "POST",
        int_refresh_min_time: int = 4000,
        max_retrieves: int = 100,
        int_status_code_ok: int = 200,
        int_status_code_iteration: int = 400,
        key_refresh_token: str = "refresh_token",
        key_access_token: str = "access_token",
        key_expires_in: str = "expires_in",
        bl_str_dict_params: bool = False,
        bl_verify: bool = False,
        i_retrieves: int = 0,
        i_aux: int = 0,
        int_expiration_time: int = 0,
        app: str = "/api/oauth/token",
    ) -> str:
        # header
        dict_headers = {
            "Authorization": "Basic {}".format(self.auth_header),
        }
        # if expiration time is inferior to base time, trigger a refresh code
        while (int_expiration_time < int_refresh_min_time) and (
            i_retrieves < max_retrieves
        ):
            #   dict_params with grant type, username, password broker code and category code -
            #       if its the first retrieve use a different dict_params dictionary
            if i_retrieves == 0:
                dict_params = {
                    "grant_type": "password",
                    "username": str(self.client_id),
                    "password": str(self.client_secret),
                    "brokerCode": str(self.broker_code),
                    "categoryCode": str(self.category_code),
                }
            else:
                dict_params = {
                    "grant_type": "refresh_token",
                    "refresh_token": refresh_token,
                }
            #   coverting to string, if its user will
            if bl_str_dict_params == True:
                dict_params = "&".join(
                    "{}={}".format(k, v) for k, v in dict_params.items()
                )
            # looping while status code not a 2xx response
            while (int_status_code_iteration != int_status_code_ok) and (
                i_aux <= max_retrieves
            ):
                try:
                    resp_req = request(
                        method=method,
                        url=self.hostname_api_line_b3 + app,
                        headers=dict_headers,
                        params=dict_params,
                        verify=bl_verify,
                    )
                except:
                    continue
                int_status_code_iteration = resp_req.status_code
                i_aux += 1
            #   raises exception when not a 2xx response
            resp_req.raise_for_status()
            #   retrieving json
            dict_token = resp_req.json()
            #   refresh token
            refresh_token = dict_token[key_refresh_token]
            #   token
            token = dict_token[key_access_token]
            #   expiration time
            int_expiration_time = dict_token[key_expires_in]
            #   iterating through number of retrieves
            i_retrieves += 1
        # return token to requester
        return token

    def app_request(
        self,
        method: str,
        app_line_b3: str,
        dict_params: Optional[Dict[str, Any]] = None,
        dict_payload: Optional[List[Dict[str, Any]]] = None,
        bl_parse_dict_params_data: bool = False,
        bl_retry_if_error: bool = False,
        bl_retry_request: bool = True,
        bl_debug_mode: bool = False,
        int_max_retrieves: int = 100,
        float_secs_sleep: Optional[float] = None,
        float_secs_sleep_increase_error: float = 1.0,
        int_status_code_ok: int = 200,
        list_int_http_error_token: List[int] = [401],
    ) -> Union[List[Dict[str, Any]], int]:
        # passing variables
        i = 0
        float_secs_sleep_iteration = float_secs_sleep
        # header
        dict_header = {
            "Authorization": "Bearer {}".format(self.token),
            "Content-Type": "application/json",
        }
        # check wheter it is needed to parse the params dictionary
        if bl_parse_dict_params_data == True:
            if dict_params != None:
                dict_params = JsonFiles().dict_to_json(dict_params)
            if dict_payload != None:
                dict_payload = JsonFiles().dict_to_json(dict_payload)
        # request instrument informations - id, symbol and asset
        if bl_retry_if_error == True:
            while (bl_retry_request == True) and (i <= int_max_retrieves):
                if bl_debug_mode == True:
                    print("*** ATTEMPT REQUEST #{} ***".format(i))
                # # request
                # print('URL: {}'.format(self.hostname_api_line_b3 + app_line_b3))
                # print('PARAMS: {}'.format(dict_params))
                # print('DATA: {}'.format(dict_payload))
                try:
                    resp_req = request(
                        method=method,
                        url=self.hostname_api_line_b3 + app_line_b3,
                        headers=dict_header,
                        params=dict_params,
                        data=dict_payload,
                    )
                    # print('ENDPOINT + API: {}'.format(resp_req.url))
                    if resp_req.status_code == int_status_code_ok:
                        bl_retry_request = False
                    elif resp_req.status_code in list_int_http_error_token:
                        #   reset token wheter http error 401 has been reached
                        dict_header = {
                            "Authorization": "Bearer {}".format(self.access_token),
                            "Content-Type": "application/json",
                        }
                    else:
                        float_secs_sleep_iteration += float_secs_sleep_increase_error
                    if bl_debug_mode == True:
                        print("REQUEST SUCCESFULLY MADE")
                except:
                    bl_retry_request = True
                    if bl_debug_mode == True:
                        print("EXCEPTION IN REQUEST #{}".format(i))
                # wait
                if float_secs_sleep_iteration != None:
                    time.sleep(float_secs_sleep_iteration)
                # iteration increase
                i += 1
            # reseting variables
            float_secs_sleep_iteration = float_secs_sleep
        else:
            resp_req = request(
                method=method,
                url=self.hostname_api_line_b3 + app_line_b3,
                headers=dict_header,
                params=dict_params,
                data=dict_payload,
            )
            if bl_debug_mode == True:
                print("REQUEST SUCCESFULLY MADE")
        #   raises exception when not a 2xx response
        resp_req.raise_for_status()
        #   retrieving response
        try:
            return resp_req.json()
        except:
            return resp_req.status_code


class Operations(ConnectionApi):

    @property
    def exchange_limits(
        self,
        method: str = "GET",
        app: str = "/api/v1.0/exchangeLimits/spxi/{}",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        return self.app_request(
            self.token,
            method,
            app.format(self.broker_code),
            bl_retry_if_error=bl_retry_if_error,
        )

    @property
    def groups_authorized_markets(
        self,
        method: str = "GET",
        app: str = "/api/v1.0/exchangeLimits/autorizedMarkets",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        return self.app_request(
            self.token, method, app, bl_retry_if_error=bl_retry_if_error
        )

    def intruments_per_group(
        self,
        group_id: str,
        method: str = "POST",
        str_bl_settled: str = "true",
        bl_parse_dict_params_data: bool = True,
        float_secs_sleep: Optional[float] = None,
        app: str = "/api/v1.0/exchangeLimits/findInstruments",
    ) -> Union[List[Dict[str, Any]], int]:
        dict_payload = {
            "authorizedMarketGroupId": group_id,
            "isLimitSetted": str_bl_settled,
        }
        return self.app_request(
            self.token,
            method,
            app,
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            float_secs_sleep=float_secs_sleep,
        )

    @property
    def authorized_markets_instruments(
        self,
        key_id: str = "id",
        key_name: str = "name",
        key_assets_associated: str = "assets_associated",
        key_limit_spci_opt: str = "limitSpciOption",
        key_limit_spvi_opt: str = "limitSpviOption",
        key_limit_spci: str = "limitSpci",
        key_limit_spvi: str = "limitSpvi",
        key_instrument_symbol: str = "instrumentSymbol",
        key_instrument_asset: str = "instrumentAsset",
    ) -> Dict[str, Union[str, int, float]]:
        # setting variables
        dict_export = dict()
        # json groups of authorized markets
        json_authorized_markets = self.groups_authorized_markets
        # loop through each authorized market and collect its assets associated
        for dict_ in json_authorized_markets:
            for dict_assets in self.intruments_per_group(dict_[key_id]):
                # print(dict_assets['instrumentSymbol'])
                #   in case the profile id is not in the exporting dict, include its as a key
                if dict_[key_id] not in dict_export:
                    dict_export[dict_[key_id]] = dict()
                #   in case it is already a key, include id, name, and assets associated in the
                #       values as a new dictionary nested
                else:
                    dict_export[dict_[key_id]][key_id] = dict_[key_id]
                    dict_export[dict_[key_id]][key_name] = dict_[key_name]
                    #   check wheter the assets associated key already exists or not, create a list if
                    #       it is not present, and appending a dictionary with symbol, asset, spci,
                    #       spvi, spci option and spvi option limits otherwise
                    if key_assets_associated not in dict_export[dict_[key_id]]:
                        dict_export[dict_[key_id]][key_assets_associated] = list()
                    else:
                        if key_limit_spci_opt in dict_assets:
                            dict_export[dict_[key_id]][key_assets_associated].append(
                                {
                                    "instrument_symbol": dict_assets[
                                        key_instrument_symbol
                                    ],
                                    "instrument_asset": dict_assets[
                                        key_instrument_asset
                                    ],
                                    "limit_spci": dict_assets[key_limit_spci],
                                    "limit_spvi": dict_assets[key_limit_spvi],
                                    "limit_spci_option": dict_assets[
                                        key_limit_spci_opt
                                    ],
                                    "limit_spvi_option": dict_assets[
                                        key_limit_spvi_opt
                                    ],
                                }
                            )
                        else:
                            dict_export[dict_[key_id]][key_assets_associated].append(
                                {
                                    "instrument_symbol": dict_assets[
                                        key_instrument_symbol
                                    ],
                                    "instrument_asset": dict_assets[
                                        key_instrument_asset
                                    ],
                                    "limit_spci": dict_assets[key_limit_spci],
                                    "limit_spvi": dict_assets[key_limit_spvi],
                                }
                            )
        #  return dictionary with ticker, id of authorized market, maximum spxi and maximum option
        #       spxi, in case it is available
        return dict_export


class Resources(Operations):

    @property
    def instrument_informations(
        self,
        method: str = "GET",
        app: str = "/api/v1.0/symbol",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        return self.app_request(
            self.token, method, app, bl_retry_if_error=bl_retry_if_error
        )

    @property
    def instrument_infos_exchange_limits(
        self,
        key_infos_id: str = "id",
        key_exchange_limits_id: str = "instrumentId",
        key_symbol: str = "symbol",
    ) -> Dict[str, Dict[Union[str, int, float]]]:
        # dataframe of exchange limits
        df_exchange_limits = pd.DataFrame.from_dict(self.exchange_limits)
        # convert data types
        df_exchange_limits = df_exchange_limits.astype({key_exchange_limits_id: str})
        # dataframe of instrument informations
        df_instrument_informations = pd.DataFrame.from_dict(
            self.instrument_informations
        )
        # convert data types
        df_instrument_informations = df_instrument_informations.astype(
            {key_infos_id: str}
        )
        # left join instrument infos and exchange limits
        df_join_instruments = df_instrument_informations.merge(
            df_exchange_limits,
            how="left",
            left_on=key_infos_id,
            right_on=key_exchange_limits_id,
        )
        # rename columns of interest
        df_join_instruments = df_join_instruments.rename(
            columns={key_symbol + "_x": key_symbol}
        )
        # remove columns of interest
        df_join_instruments.drop(columns=[key_symbol + "_y"], inplace=True)
        # exporting dictionary with symbol as key
        return {
            row[key_symbol]: {col_: row[col_] for col_ in df_join_instruments.columns}
            for _, row in df_join_instruments.iterrows()
        }

    def instrument_id_by_symbol(
        self, symbol: str, method: str = "GET", app: str = "/api/v1.0/symbol/{}"
    ) -> Union[List[Dict[str, Any]], int]:
        return self.app_request(self.token, method, app.format(str(symbol)))


class AccountsData(ConnectionApi):

    def client_infos(
        self,
        account_code: str,
        bl_retry_if_error: bool = True,
        bl_debug_mode: bool = False,
        float_secs_sleep: Optional[float] = None,
        method: str = "GET",
        app: str = "/api/v1.0/account",
    ) -> Union[List[Dict[str, Any]], int]:
        # parameters
        dict_params = {
            "participantCode": self.broker_code,
            "pnpCode": self.category_code,
            "accountCode": account_code,
        }
        # retrieving json
        return self.app_request(
            self.token,
            method,
            app,
            dict_params=dict_params,
            bl_retry_if_error=bl_retry_if_error,
            float_secs_sleep=float_secs_sleep,
            bl_debug_mode=bl_debug_mode,
        )

    def spxi_get(
        self,
        account_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/account/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        # parameters
        dict_params = {
            "accId": account_id,
        }
        #   retrieving json
        return self.app_request(self.token, method, app.format(account_id), dict_params)

    def spxi_instrument_post(
        self,
        account_id: str,
        dict_payload: Union[List[Dict[str, Any]], None],
        bl_parse_dict_params_data: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/account/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        SPXI inclustion to account
        Args:
            account_id (str): account id
            dict_payload (dict):
                [{'instrumentId': int, 'isRemoved': 'false', 'spci': int, 'spvi': int, 'symbol': str}, ...]
            bl_parse_dict_params_data (bool): parse dictionary params data
            method (str): method
            app (str): app
        INPUTS: ACCOUNT ID, DICTIONARY PAYLOAD (LIST OF DICTIONARIES WITH KEYS: 'instrumentId': int,
            'isRemoved': 'false', 'spci': int, 'spvi': int, 'symbol': str), BOOLEAN PARSE DICT,
            METHOD (DEFAULT), APP SPXI (DEFAULT)
        OUTPUTS: STATUS CODE
        """
        return self.app_request(
            self.token,
            method,
            app.format(account_id),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
        )

    def spxi_instrument_delete(
        self,
        account_id: str,
        dict_payload: Union[List[Dict[str, Any]], None],
        bl_parse_dict_params_data=True,
        method="POST",
        app="/api/v1.0/account/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        SPXI instrument removal
        Args:
            account_id (str): account id
            dict_payload (dict):
                [{'instrumentId': int, 'isRemoved': 'false', 'symbol': str}, ...]
            bl_parse_dict_params_data (bool): parse dictionary params data
            method (str): method
            app (str): app
        INPUTS: ACCOUNT ID, DICTIONARY PAYLOAD (LIST OF DICTIONARIES WITH KEYS: 'instrumentId': int,
            'isRemoved': 'false', 'spci': int, 'spvi': int, 'symbol': str), BOOLEAN PARSE DICT,
            METHOD (DEFAULT), APP SPXI (DEFAULT)
        OUTPUTS: STATUS CODE
        """
        return self.app_request(
            self.token,
            method,
            app.format(account_id),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
        )

    def spxi_tmox_global_metrics_remove(
        self,
        account_id: str,
        method: str = "DELETE",
        app: str = "/api/v1.0/account/{}/lmt",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # parameters
        dict_params = {
            "accId": account_id,
        }
        # retrieving json
        return self.app_request(self.token, method, app.format(account_id), dict_params)

    def specific_global_metric_remotion(
        self,
        account_id: str,
        metric: str,
        method: str = "DELETE",
        app: str = "/api/v2.0/account/{}/lmt",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # parameters
        dict_params = {"accId": account_id, "metric": metric}
        # retrieving json
        return self.app_request(self.token, method, app.format(account_id), dict_params)


class DocumentsData(ConnectionApi):

    def doc_info(
        self,
        doc_code: str,
        bl_retry_if_error: bool = True,
        method: str = "GET",
        app: str = "/api/v1.0/document",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET INFOS REGARDING DOCUMENT CODE (HIGLIGHT TO THE POSSIBILITY TO RETRIEVE
            DOCUMENT ID FROM DOCUMENT CODE)
        INPUTS: DOCUMENT CODE, BL RETRY IF ERROR (DEFAULT), METHOD (DEFAULT), APP (DEFAULT)
        OUTPUTS: JSON
        """
        # payload
        dict_params = {
            "participantCode": self.broker_code,
            "pnpCode": self.category_code,
            "documentCode": str(doc_code),
        }
        # retrieving json
        return self.app_request(
            self.token,
            method,
            app,
            dict_params=dict_params,
            bl_retry_if_error=bl_retry_if_error,
        )

    def block_unblock_doc(
        self,
        doc_id: str,
        bl_isblocked: bool = True,
        bl_parse_dict_params_data: bool = True,
        bl_retry_if_error: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/document/{}",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: BLOCK DOCUMENT
        INPUTS: DOC_ID, BL_ISBLOCKED
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        # payload
        dict_params = {"id": str(doc_id), "isBlocked": bl_isblocked}
        # retrieving json
        return self.app_request(
            self.token,
            method,
            app.format(str(doc_id)),
            dict_params=dict_params,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            bl_retry_if_error=bl_retry_if_error,
        )

    def update_profile(
        self,
        doc_id: str,
        doc_profile_id: str,
        bl_parse_dict_params_data: bool = True,
        int_rmkt_evaluation: int = 0,
        bl_retry_if_error: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/document/{}",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: UPDATE DOCUMENT PROFILE
        INPUTS: DOC ID, DOC PROFILE ID, RMKT EVALUATION (DEFAULT)
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        # payload
        dict_payload = {
            "id": str(doc_id),
            "profileFull": int(doc_profile_id),
            "rmktEvaluation": int_rmkt_evaluation,
        }
        #   retrieving json
        return self.app_request(
            self.token,
            method,
            app.format(str(doc_id)),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            bl_retry_if_error=bl_retry_if_error,
        )

    def bl_protection_mode(
        self,
        doc_id: str,
        bl_protect: bool = True,
        bl_parse_dict_params_data: bool = True,
        bl_retry_if_error: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/document/{}",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: PROTECTION MODE FOR THE CURRENT DOCUMET
        INPUTS: DOC ID, DOC PROFILE ID, RMKT EVALUATION (DEFAULT)
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        # payload
        dict_payload = {
            "id": str(doc_id),
            "isProtected": str(bl_protect).lower(),
        }
        #   retrieving json
        return self.app_request(
            self.token,
            method,
            app.format(str(doc_id)),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            bl_retry_if_error=bl_retry_if_error,
        )

    def client_infos(
        self,
        doc_id: str,
        bl_retry_if_error: bool = True,
        float_secs_sleep: Optional[float] = None,
        bl_debug_mode: bool = False,
        method: str = "GET",
        app: str = "/api/v1.0/account",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: CLIENT REGISTER ON LINE B3
        INPUTS: DOC ID, METHOD (DEFAULT), APP CLIENT INFOS (DEFAULT)
        OUTPUTS: JSON - KEYS: INTERNAL ID, CODE, DOCUMENT, BOOLEAN PROTECTED, BOOLEAN BLOCKED,
            NAME, CATEGORY, SEGMENT, STATUS, PARTICIPANT NAME, PNP NAME, PARTICIPANT CODE,
            PNP CODE, PNP ACCOUNT CODE, PARTICIPANT ACCOUNT CODE, PARTICIPANT ACCOUNT CODE,
            ACCOUNT TYPE, OWNER DOCUMENT CODE, OWNER NAME
        """
        # parameters
        dict_params = {
            "participantCode": self.broker_code,
            "pnpCode": self.category_code,
            "documentId": doc_id,
        }
        # retrieving json
        return self.app_request(
            self.token,
            method,
            app,
            dict_params=dict_params,
            bl_retry_if_error=bl_retry_if_error,
            float_secs_sleep=float_secs_sleep,
            bl_debug_mode=bl_debug_mode,
        )

    def doc_profile(
        self,
        doc_id: str,
        method: str = "GET",
        app: str = "/api/v2.0/document/v2.0/document/{}",
        key_api_line_b3_profile_full: str = "profileFull",
        key_api_line_b3_profile_name: str = "profileName",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: DOC PROFILE (INTEGER AND NAME)
        INPUTS: DOC ID, METHOD (DEFAULT), APP CLIENT INFOS (DEFAULT)
        OUTPUTS: INTEGER AND STRIN
        """
        #   fetch json
        json_doc = self.app_request(
            self.token, method, app.format(doc_id), bl_retry_if_error=bl_retry_if_error
        )
        #   returning profile info
        return {
            "profile_id": json_doc[key_api_line_b3_profile_full],
            "profile_name": json_doc[key_api_line_b3_profile_name],
        }

    def spxi_get(
        self,
        doc_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/document/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET SPXI INFORMATION FROM CLIENT DOCUMENT
        INPUTS: DOC ID, METHOD, APP SPXI, KEY ID (FROM CLIENT INFOS)
        OUTPUTS: JSON - KEYS: INSTRUMENT ID, SYMBOL, SPCI, SPVI, SPCI EXCHANGE, SPVI EXCHANGE,
            IS REMOVED
        """
        # parameters
        dict_params = {
            "docId": doc_id,
        }
        #   retrieving json
        return self.app_request(self.token, method, app.format(doc_id), dict_params)

    def spxi_instrument_post(
        self,
        doc_id: str,
        dict_payload: Union[List[Dict[str, Any]], None],
        bl_parse_dict_params_data=True,
        method="POST",
        app="/api/v1.0/document/{}/lmt/spxi",
        bl_retry_if_error=True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: SPXI INCLUSION TO DOCUMENT
        INPUTS: DOCUMENT ID, DICTIONARY PAYLOAD (LIST OF DICTIONARIES WITH KEYS: 'instrumentId': int,
            'isRemoved': 'false', 'spci': int, 'spvi': int, 'symbol': str), BOOLEAN PARSE DICT,
            METHOD (DEFAULT), APP SPXI (DEFAULT)
        OUTPUTS: STATUS CODE
        """
        return self.app_request(
            self.token,
            method,
            app.format(doc_id),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            bl_retry_if_error=bl_retry_if_error,
        )

    def spxi_instrument_delete(
        self,
        doc_id: str,
        dict_payload: Union[List[Dict[str, Any]], None],
        bl_parse_dict_params_data: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/document/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: SPXI INCLUSION TO DOCUMENT
        INPUTS: DOCUMENT ID, DICTIONARY PAYLOAD (LIST OF DICTIONARIES WITH KEYS: 'instrumentId': int,
            'isRemoved': 'true', 'symbol': str), BOOLEAN PARSE DICT, METHOD (DEFAULT),
            APP SPXI (DEFAULT)
        OUTPUTS: STATUS CODE
        """
        return self.app_request(
            self.token,
            method,
            app.format(doc_id),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
        )


class Professional(ConnectionApi):

    def professional_code_get(
        self,
        method: str = "GET",
        app: str = "/api/v1.0/operationsProfessionalParticipant/code",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFESSIONAL INFORMATION FROM ITS CODE
        INPUTS: PROFESSIONAL CODE
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        dict_params = {
            "participantCode": self.broker_code,
            "pnpCode": self.category_code,
        }
        return self.app_request(self.token, method, app, dict_params)

    def professional_historic_position(
        self,
        professional_code: str,
        dt_start: str,
        dt_end: str,
        int_participant_perspective_type: int = 0,
        list_metric_type: List[int] = [
            1,
            2,
            3,
            4,
            6,
            7,
            22,
            25,
            26,
            27,
            28,
            29,
            36,
            38,
            39,
        ],
        entity_type: int = 4,
        int_items_per_page: int = 50,
        method: str = "POST",
        app: str = "https://api.line.trd.cert.bvmfnet.com.br/api/v2.0/position/hstry",
        bl_retry_if_error: bool = True,
        bl_debug_mode: bool = True,
        bl_parse_dict_params_data: bool = True,
        float_secs_sleep: Optional[float] = None,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFESSIONAL POSITIONS HISTORIC FROM ITS CODE
        INPUTS: PROFESSIONAL CODE
        OUTPUTS: JSON WITH POSITIONS HISTORIC
        """
        # payload for request
        dict_payload = {
            "angularItensPerPage": int_items_per_page,
            "entityType": entity_type,
            "metricTypes": list_metric_type,
            "ownerBrokerCode": int(self.broker_code),
            "ownerCategoryType": int(self.category_code),
            "partPerspecType": int_participant_perspective_type,
            "registryDateEnd": dt_end,
            "registryDateStart": dt_start,
            "traderCode": professional_code,
        }
        if bl_debug_mode == True:
            pprint(dict_payload)
        # retrieving professional positions
        return self.app_request(
            self.token,
            method,
            app,
            dict_payload=dict_payload,
            bl_retry_if_error=bl_retry_if_error,
            bl_debug_mode=bl_debug_mode,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            float_secs_sleep=float_secs_sleep,
        )


class ProfilesData(ConnectionApi):

    @property
    def risk_profile(
        self, method: str = "GET", app: str = "/api/v1.0/riskProfile"
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFILES AVAILABLE IN LINE B3
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        return self.app_request(self.token, method, app)

    def entities_associated_profile(
        self,
        id_profile: str,
        method: str = "GET",
        app: str = "/api/v1.0/riskProfile/enty",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: ENTITY DOCUMENTS LINKED TO THE PROFILE
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        dict_params = {
            "id": id_profile,
            "participantCode": self.broker_code,
            "pnpCode": self.category_code,
        }
        return self.app_request(
            self.token, method, app, dict_params, bl_retry_if_error=bl_retry_if_error
        )

    def profile_global_limits_get(
        self,
        prof_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/riskProfile/{}/lmt",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFILE GLOBAL LIMITS BY ITS ID
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        return self.app_request(self.token, method, app.format(prof_id))

    def profile_market_limits_get(
        self,
        prof_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/riskProfile/{}/lmt/mkta",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFILE MARKET LIMITS BY ITS ID
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        return self.app_request(
            self.token, method, app.format(prof_id), bl_retry_if_error=bl_retry_if_error
        )

    def profile_spxi_limits_get(
        self,
        prof_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/riskProfile/{}/lmt/spxi",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFILE SPCI/SPVI LIMITS BY ITS ID
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        return self.app_request(self.token, method, app.format(prof_id))

    def profile_tmox_limits_get(
        self,
        prof_id: str,
        method: str = "GET",
        app: str = "/api/v1.0/riskProfile/{}/lmt/tmox",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: GET PROFILE TMOC/TMOV LIMITS BY ITS ID
        INPUTS: PROFILE ID
        OUTPUTS: JSON WITH ID, NAME, BLOCKED/PROTECTED, PROFILE ID/NAME AND PROFESSIONAL CODE
        """
        return self.app_request(self.token, method, app.format(prof_id))

    def spxi_instrument_post(
        self,
        prof_id: str,
        dict_payload: List[Dict[str, Any]],
        bl_parse_dict_params_data: bool = True,
        method: str = "POST",
        app: str = "/api/v1.0/riskProfile/{}/lmt/tmox",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: SPXI INCLUSION TO DOCUMENT
        INPUTS: DOCUMENT ID, DICTIONARY PAYLOAD (LIST OF DICTIONARIES WITH KEYS: 'instrumentId': int,
            'isRemoved': false, 'symbol': 'RNEW4', 'tmoc': 0, 'tmocExchange': 0, 'tmov': 1000000,
            'tmovExchange': 0), BOOLEAN PARSE DICT, METHOD (DEFAULT), APP SPXI (DEFAULT)
        OUTPUTS: STATUS CODE
        """
        return self.app_request(
            self.token,
            method,
            app.format(prof_id),
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            bl_retry_if_error=bl_retry_if_error,
        )


class Monitoring(ConnectionApi):

    @property
    def alerts(
        self,
        method: str = "GET",
        app: str = "/api/v1.0/alert/lastalerts?filterRead=true",
        bl_retry_if_error: bool = True,
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: INSTRUMENTS INFORMATION
        INPUTS: MEHTOD (DEFAULT), APP INSTRUMENT INFORMATION (DEFAULT)
        OUTPUTS: JSON (ID, SYMBOL AND ASSET)
        """
        return self.app_request(
            self.token, method, app, bl_retry_if_error=bl_retry_if_error
        )


class SystemEventManagement(ConnectionApi):

    def report(
        self,
        int_working_days_before: int = 1,
        int_working_days_after: int = 0,
        str_start_time: str = "00:00",
        str_sup_time: str = "23:59",
        str_null: str = "null",
        int_entity_type: int = 3,
        method: str = "POST",
        bl_parse_dict_params_data: str = True,
        float_secs_sleep: Optional[float] = None,
        app: str = "/api/v1.0/systemEvent",
    ) -> Union[List[Dict[str, Any]], int]:
        """
        DOCSTRING: SPXI INCLUSION TO DOCUMENT
        INPUTS: WORKING DAYS BEFORE AND AFTER
        OUTPUTS: STATUS CODE
        """
        # payload to consult the range of dates of interest
        dict_payload = {
            "participantCode": int(self.broker_code),
            "categoryType": int(self.category_code),
            "entityType": int_entity_type,
            "carryingAccountCode": str_null,
            "pnpCode": "",
            "accountTypeLineDomain": str_null,
            "ownerName": str_null,
            "documentCode": str_null,
            "accountCode": str_null,
            "startTime": str_start_time,
            "endTime": str_sup_time,
            "startDate": DatesBR()
            .sub_working_days(DatesBR().curr_date(), int_working_days_before)
            .strftime("%d/%m/%Y"),
            "endDate": DatesBR()
            .add_working_days(DatesBR().curr_date(), int_working_days_after)
            .strftime("%d/%m/%Y"),
        }
        # return json
        return self.app_request(
            self.token,
            method,
            app,
            dict_payload=dict_payload,
            bl_parse_dict_params_data=bl_parse_dict_params_data,
            float_secs_sleep=float_secs_sleep,
        )



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class HalfSubtractor(metaclass=TypeChecker):

    def __init__(self, a: int, b: int) -> None:
        """
        Initialize the Half Subtractor with two binary inputs.
        Args:
            a (int): First binary input (0 or 1)
            b (int): Second binary input (0 or 1)
        Returns:
            None
        """
        self.a = a
        self.b = b

    @property
    def difference(self) -> int:
        """
        Returns the difference output of the half subtractor.
        The difference output indicates the difference of the two binary inputs.
        The difference is calculated using the XOR operation:
            - Difference (D) = A XOR B
        Args:
            None
        Returns:
            int: The difference output of the half subtractor.
        """
        return self.a ^ self.b

    @property
    def borrow(self) -> int:
        """
        Returns the borrow output of the half subtractor.
        The borrow output indicates whether there is a borrow from the subtraction
        of the two binary inputs.
        The borrow is calculated using the following logic:
            - Borrow occurs when A is less than B
        Args:
            None
        Returns:
            int: The borrow output of the half subtractor.
        """
        return not self.a and self.b


class FullSubtractor(metaclass=TypeChecker):

    def __init__(self, a: int, b: int, borrow_in: int) -> None:
        """
        Initialize the Full Subtractor with two binary inputs and a borrow-in.
        Args:
            a (int): First binary input (0 or 1)
            b (int): Second binary input (0 or 1)
            borrow_in (int): Borrow input from the previous stage (0 or 1).
        Returns:
            None
        """
        self.a = a
        self.b = b
        self.borrow_in = borrow_in

    @property
    def difference(self) -> int:
        """
        Returns the difference output of the full subtractor.
        The difference output indicates the difference of the two binary inputs and
        the borrow-in. The difference is calculated using the XOR operation:
            - Difference (D) = A XOR B XOR Borrow-In
        Args:
            None
        Returns:
            int: The difference output of the full subtractor.
        """
        return (self.a ^ self.b) ^ self.borrow_in

    @property
    def borrow_out(self) -> int:
        """
        Returns the borrow output of the full subtractor.
        The borrow output indicates whether there is a borrow from the subtraction
        of the two binary inputs and the borrow-in.
        The borrow is calculated using the following logic:
            - Borrow occurs when A is less than B, or
            - Borrow occurs when Borrow-In is 1 and A XOR B is 0
        Args:
            None
        Returns:
            int: The borrow output of the full subtractor.
        """
        return (not self.a and self.b) or (self.borrow_in and not (self.a ^ self.b))


if __name__ == "__main__":
    hs = HalfSubtractor(1, 1)
    # output: Difference: 0, Borrow: 1
    print(f"Half Subtractor (1, 1) => Difference: {hs.difference()}, Borrow: {hs.borrow()}")

    fs = FullSubtractor(1, 0, 1)
    # output: Difference: 0, Borrow Out: 1
    print(f"Full Subtractor (1, 0, 1) => Difference: {fs.difference()}, Borrow Out: {fs.borrow_out()}")



import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Tuple
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from selenium.webdriver.remote.webdriver import WebDriver
from stpstone._config.global_slots import YAML_WW_TRADING_ECON
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD


class TradingEconWW(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_WW_TRADING_ECON,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs

    def list_web_elements(self, cls_selenium_wd: SeleniumWD, web_driver: WebDriver, xpath_: str) \
        -> List[Any]:
        """
        Get list of web elements from xpath

        Args:
            cls_selenium_wd (SeleniumWD): selenium web web_driver
            web_driver (WebDriver): selenium web web_driver
            xpath (str): xpath to find elements

        Returns:
            list: list of contents of web elements
        """
        list_els = cls_selenium_wd.find_elements(web_driver, xpath_)
        list_els = [x.text for x in list_els]
        return list_els

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        source = self.get_query_params(resp_req.url, "source")
        try:
            cls_selenium = SeleniumWD(resp_req.url, bl_headless=True, bl_incognito=True)
            web_driver = cls_selenium.get_web_driver
            list_th = list(YAML_WW_TRADING_ECON[source]["dtypes"].keys())
            list_td = self.list_web_elements(cls_selenium, web_driver, 
                                             YAML_WW_TRADING_ECON[source]["xpaths"]["list_td"])
        finally:
            web_driver.quit()
        list_ser = HandlingDicts().pair_headers_with_data(list_th, list_td)
        return pd.DataFrame(list_ser)



import MetaTrader5 as mt5
import time
import pandas as pd
from datetime import datetime
from math import ceil


class MT5():

    def package_info(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        print("MetaTrader5 package author: ", mt5.__author__)
        print("MetaTrader5 package version: ", mt5.__version__)

    def initialize(self, path, login, server, password):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        if not mt5.initialize(path=path, login=login, server=server, password=password):
            print("initialize() failed, error code =", mt5.last_error())
            # shut down connection to the MetaTrader 5 terminal
            mt5.shutdown()
            # quit()

    def shutdown(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        mt5.shutdown()

    def symbols_get(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return mt5.symbols_get()

    def symbols_total(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        symbols = mt5.symbols_total()
        if symbols > 0:
            print("Total symbols =", symbols)
            return symbols
        else:
            print("Symbols not found")
            return None

    def get_symbols_info(self, market_data=True):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        symbols = mt5.symbols_get()
        dict_tickers = dict()
        if market_data:
            qtd_vezes = ceil(len(symbols) / 4900)
            l = 0
            for j in range(qtd_vezes):
                lim_inf = j * 4900
                if j == qtd_vezes - 1:
                    lim_sup = len(symbols)
                else:
                    lim_sup = (j + 1) * 4900
                print('Limite inferior:', lim_inf)
                print('Limite superior:', lim_sup)
                for symbol in symbols[lim_inf:lim_sup]:
                    mt5.symbol_select(symbol.name, True)
                print('Aguardando')
                time.sleep(10)
                for symbol in symbols[lim_inf:lim_sup]:
                    ticker_info = mt5.symbol_info(symbol.name)
                    if ticker_info != None:
                        dict_tickers[l] = ticker_info._asdict()
                        l += 1
                for symbol in symbols[lim_inf:lim_sup]:
                    mt5.symbol_select(symbol.name, False)
        else:
            l = 0
            for symbol in symbols:
                dict_tickers[l] = symbol._asdict()
                l += 1
        df = pd.DataFrame(dict_tickers).T
        df.loc[:, 'paths'] = df.path.str.split('\\')
        df.loc[:, 'CLASSE1'] = df.apply(lambda row: row['paths'][0], axis=1)

        def tipo2(row):
            if len(row.paths) == 2:
                return 'BMF'
            else:
                return row.paths[1]

        def tipo3(row):
            if len(row.paths) == 2:
                return row.paths[1]
            else:
                return row.paths[2]
        df.loc[:, 'CLASSE2'] = df.apply(tipo2, axis=1)
        df.loc[:, 'CLASSE3'] = df.apply(tipo3, axis=1)

        def exp_time(row):
            try:
                result = pd.to_datetime(row.expiration_time, unit='s')
            except:
                result = pd.to_datetime(row.expiration_time, unit='ms')
            return result
        df.loc[:, 'expiration_time'] = df.apply(exp_time, axis=1)
        df.loc[:, 'time'] = pd.to_datetime(df['time'], unit='s')
        # symbols=mt5.symbols_get()
        if df.shape[0] > 0:
            print("Number of symbols =", df.shape[0])
            return df
        else:
            print("Symbols not found")
            return None

    def get_all_info_of_symbols(self, symbols):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        dict_tickers = dict()
        i = 0
        for symbol in symbols:
            dict_tickers[i] = symbol._asdict()
            i += 1
        df = pd.DataFrame(dict_tickers).T
        df.loc[:, 'paths'] = df.path.str.split('\\')
        df.loc[:, 'CLASSE1'] = df.apply(lambda row: row['paths'][0], axis=1)

        def tipo2(row):
            if len(row.paths) == 2:
                return 'BMF'
            else:
                return row.paths[1]

        def tipo3(row):
            if len(row.paths) == 2:
                return row.paths[1]
            else:
                return row.paths[2]
        df.loc[:, 'CLASSE2'] = df.apply(tipo2, axis=1)
        df.loc[:, 'CLASSE3'] = df.apply(tipo3, axis=1)
        df.loc[:, 'expiration_time'] = pd.to_datetime(
            df['expiration_time'], unit='s')
        df.loc[:, 'time'] = pd.to_datetime(df['time'], unit='s')
        print('Total de tickers: {}'.format(df.shape[0]))
        return df

    def get_ticks_from(self, symbol, datetime_from, ticks_qty,
                       type_ticks=mt5.COPY_TICKS_ALL, print_quantidade=False):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        if not isinstance(datetime_from, datetime):
            print('datetime_from is not a datetime object')
            return None
        ticks = mt5.copy_ticks_from(
            symbol, datetime_from, ticks_qty, type_ticks)
        # a partir dos dados recebidos criamos o DataFrame
        ticks_frame = pd.DataFrame(ticks)
        if print_quantidade:
            print("Ticks recebidos:", ticks_frame.shape[0])
        if not ticks_frame.empty:
            # convertemos o tempo em segundos no formato datetime
            ticks_frame['time'] = pd.to_datetime(ticks_frame['time'], unit='s')
            ticks_frame['time_msc'] = pd.to_datetime(
                ticks_frame['time_msc'], unit='ms')
        return ticks_frame

    def get_ticks_range(self, symbol, datetime_from, datetime_to,
                        type_ticks=mt5.COPY_TICKS_ALL, print_quantidade=False):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        if not isinstance(datetime_from, datetime) or not isinstance(datetime_to, datetime):
            print('datetime_from or datetime_to is not a datetime object')
            return None
        ticks = mt5.copy_ticks_range(
            symbol, datetime_from, datetime_to, type_ticks)
        # a partir dos dados recebidos criamos o DataFrame
        ticks_frame = pd.DataFrame(ticks)
        if print_quantidade:
            print("Ticks recebidos:", ticks_frame.shape[0])
        if not ticks_frame.empty:
            # convertemos o tempo em segundos no formato datetime
            ticks_frame['time'] = pd.to_datetime(ticks_frame['time'], unit='s')
            ticks_frame['time_msc'] = pd.to_datetime(
                ticks_frame['time_msc'], unit='ms')
        return ticks_frame

    def get_last_tick(self, symbol):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        last_tick = mt5.symbol_info_tick(symbol)
        if last_tick:
            return last_tick
        else:
            print("mt5.symbol_info_tick({}) failed, error code =".format(
                symbol), mt5.last_error())
            return None

    def get_market_depth(self, ticker, n_times=10):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        if mt5.market_book_add(ticker):
            # get the market depth data 10 times in a loop
            for i in range(n_times):
                # get the market depth content (Depth of Market)
                items = mt5.market_book_get(ticker)
                # display the entire market depth 'as is' in a single string
                print(items)
                # now display each order separately for more clarity
                if items:
                    for it in items:
                        # order content
                        print(it._asdict())
                # pause for 5 seconds before the next request of the market depth data
                time.sleep(5)
            # cancel the subscription to the market depth updates (Depth of Market)
            mt5.market_book_release(ticker)
        else:
            print("mt5.market_book_add({}) failed, error code =".format(
                ticker), mt5.last_error())
        # shut down connection to the MetaTrader 5 terminal
        return items

    def enable_display_marketwatch(self, ticker):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        selected = mt5.symbol_select(ticker, True)
        if not selected:
            print("Failed to select {}".format(ticker))

    def get_symbol_info_tick(self, ticker):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # display the last GBPUSD tick
        lasttick = mt5.symbol_info_tick(ticker)
        print(lasttick)
        # display tick field values in the form of a list
        print("Show symbol_info_tick({})._asdict():".format(ticker))
        symbol_info_tick_dict = mt5.symbol_info_tick(ticker)._asdict()
        for prop in symbol_info_tick_dict:
            print("  {}={}".format(prop, symbol_info_tick_dict[prop]))
        return symbol_info_tick_dict

    def get_symbol_properties(self, ticker):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # display EURJPY symbol properties
        symbol_info = mt5.symbol_info(ticker)
        if symbol_info != None:
            # display the terminal data 'as is'
            print(symbol_info)
            print("{}: spread = {} digits = {}".format(
                ticker, symbol_info.spread))
            # display symbol properties as a list
            print("Show symbol_info({})._asdict():".format(ticker))
            symbol_info_dict = mt5.symbol_info(ticker)._asdict()
            for prop in symbol_info_dict:
                print("  {}={}".format(prop, symbol_info_dict[prop]))
        return symbol_info_dict



from typing import Union, Dict, List, Optional
from logging import Logger
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD
from stpstone.utils.connections.netops.proxies.abc import ABCSession
from stpstone.utils.geography.ww import WWTimezones, WWGeography


class ProxyNova(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )
        self.fstr_url = "https://www.proxynova.com/proxy-server-list/country-{}/"
        self.xpath_tr = '//*[@id="tbl_proxy_list"]/tbody/tr'
        self.url = self.fstr_url.format(str_country_code.lower())
        self.selenium_wd = SeleniumWD(
            url=self.url,
            bl_headless=True
        )
        self.driver = self.selenium_wd.get_web_driver

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        list_ser = list()
        el_trs = self.selenium_wd.find_elements(self.driver, self.xpath_tr)
        for el_tr in el_trs:
            ip = self.selenium_wd.find_element(el_tr, './td[1]').text
            port = self.selenium_wd.find_element(el_tr, './td[2]').text
            last_checked = self.selenium_wd.find_element(el_tr, './td[3]').text
            proxy_speed = self.selenium_wd.find_element(el_tr, './td[4]').text
            uptime_tested_times = str(self.selenium_wd.find_element(el_tr, './td[5]').text)
            uptime = float(uptime_tested_times.split("(")[0].replace("%", "")) / 100.0
            int_times_alive = int(uptime_tested_times.split("(")[1].replace(")", ""))
            country_city = self.selenium_wd.find_element(el_tr, './td[6]').text
            country = country_city.split(" - ")[0].strip()
            city = country_city.split("- ")[1].strip()
            anonymity = self.selenium_wd.find_element(el_tr, './td[7]').text
            list_ser.append({
                "protocol": "http",
                "bl_alive": True,
                "status": "success",
                "alive_since": self.time_ago_to_ts_unix(last_checked),
                "anonymity": anonymity.lower(),
                "average_timeout": 1.0 / self.proxy_speed_to_float(proxy_speed),
                "first_seen": self.time_ago_to_ts_unix(last_checked),
                "ip_data": "",
                "ip_name": "",
                "timezone": ", ".join(WWTimezones().get_timezones_by_country_code(
                    self.str_country_code)),
                "continent": WWGeography().get_continent_by_country_code(self.str_country_code),
                "continent_code": WWGeography().get_continent_code_by_country_code(
                    self.str_country_code).upper(),
                "country": country,
                "country_code": self.str_country_code.upper(),
                "city": city,
                "district": "",
                "region_name": "",
                "zip": "",
                "bl_hosting": True,
                "isp": "",
                "latitude": 0.0,
                "longitude": 0.0,
                "organization": "",
                "proxy": True,
                "ip": ip,
                "port": port,
                "bl_ssl": True,
                "timeout": 1.0 / self.proxy_speed_to_float(proxy_speed),
                "times_alive": int_times_alive,
                "times_dead": "",
                "ratio_times_alive_dead": 1.0,
                "uptime": uptime
            })
        return list_ser



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class Utilities(metaclass=TypeChecker):

    def bitwise_and(self, a: int, b: int) -> int:
        return a & b

    def bitwise_or(self, a: int, b: int) -> int:
        return a | b

    def bitwise_xor(self, a: int, b: int) -> int:
        return a ^ b

    def bitwise_not(self, a: int) -> int:
        return ~a


if __name__ == "__main__":
    utils = Utilities()
    print(f"Bitwise AND (3, 5) => {utils.bitwise_and(3, 5)}")  # output: 1
    print(f"Bitwise OR (3, 5) => {utils.bitwise_or(3, 5)}")    # output: 7
    print(f"Bitwise XOR (3, 5) => {utils.bitwise_xor(3, 5)}")  # output: 6
    print(f"Bitwise NOT (3) => {utils.bitwise_not(3)}")        # output: -4



import pandas as pd
from datetime import datetime
from typing import Optional, List
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from urllib.parse import urlparse
from stpstone._config.global_slots import YAML_BR_PTAX_BCB
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler


class PTAXBCB(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_start:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 2),
        dt_end:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None,
        list_slugs:Optional[List[str]]=None,
        bl_debug:bool=False
    ) -> None:
        super().__init__(
            dict_metadata=YAML_BR_PTAX_BCB,
            session=session,
            cls_db=cls_db,
            logger=logger,
            token=token
        )
        self.session = session
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.bl_debug = bl_debug
        self.dt_sup_yyyymmdd = self.dt_end.strftime('%Y%m%d')
        self.dt_inf_ddmmyyyy = self.dt_start.strftime('%d/%m/%Y')
        self.dt_sup_ddmmyyyy = self.dt_end.strftime('%d/%m/%Y')
        self.df_ids = self.source('ids', bl_fetch=True)
        self.list_slugs = list_slugs if list_slugs is not None else \
            list(self.df_ids['CURRENCY_ID'].unique())

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        tup_parsed_url = urlparse(resp_req.url)
        if (f'{tup_parsed_url.scheme}://{tup_parsed_url.hostname}/'
             == YAML_BR_PTAX_BCB['ids']['host']) \
            and ('&' not in tup_parsed_url.query):
            root = HtmlHandler().lxml_parser(resp_req)
            selectors_currency = HtmlHandler().lxml_xpath(
                root, YAML_BR_PTAX_BCB['ids']['xpaths']['currency_options']
            )
            list_ser = [
                {
                    'CURRENCY_ID': selector.get('value'),
                    'CURRENCY_NAME': selector.text
                } for selector in selectors_currency
            ]
            return pd.DataFrame(list_ser)
        else:
            return None

    @property
    def composition_currency(self) -> pd.DataFrame:
        """
        Composition of available currency rates and close currency rates.
        - fetches the available currency rates from PTAX BCB;
        - left merge with the ids dataframe to add the CURRENCY_ID;
        - fetches the close currency rates, and performs left-join to add CURRENCY_NAME,
            COUNTRY_CODE and COUNTRY_NAME

        Args:
            None

        Returns:
            pd.DataFrame

        Metadata: https://www.bcb.gov.br/estabilidadefinanceira/historicocotacoes
        """
        df_available_currencies = super().source(
            'available_currencies', bl_fetch=True, bl_debug=self.bl_debug)
        df_available_currencies = df_available_currencies.merge(
            self.df_ids,
            how='left',
            left_on='NAME',
            right_on='CURRENCY_NAME',
            suffixes=('_', '')
        )
        df_close_currency_rates = super().source(
            'close_currency_rates', bl_fetch=True, bl_debug=self.bl_debug)
        df_close_currency_rates = df_close_currency_rates.merge(
            df_available_currencies,
            how='left',
            left_on='CURRENCY_SYMBOL',
            right_on='SYMBOL',
            suffixes=('_', '')
        )
        return df_close_currency_rates



### MODULE TO CONNECT IN DATABRICKS SQL ###

import sys
import os
import pyodbc as pyo
import pandas as pd
from func_timeout import func_timeout, FunctionTimedOut
sys.path.append('\\'.join([d for d in os.path.dirname(
    os.path.realpath(__file__)).split('\\')][:-2]))
from stpstone.utils.loggs.create_logs import CreateLog


class Databricks:

    def __init__(self, str_query, list_dsns, logger=None, max_error_attempts=3):
        self.str_query = str_query
        self.list_dsns = list_dsns
        self.logger = logger
        self.max_error_attempts = max_error_attempts

    def conn_dsn_databricks(self, dsn_conn, int_timeout=108000, bl_autocommit=True):
        """
        DOCSTRING: ESTABLISH CONNECTION WITH DATABRICKS
        INPUTS: CONNECTION STRING
        OUTPUTS: OBJECT WITH DATABRICKS CONNECTION
        """
        return pyo.connect('DSN={}'.format(dsn_conn), autocommit=bl_autocommit, timeout=int_timeout)

    def fetch_data_from_databricks(self, connection):
        """
        DOCSTRING: ATTEMPT TO CONSULT DATABRICKS
        INPUTS: QUERY AND CONNECTION OBJECT
        OUTPUTS: PANDAS DATAFRAME
        """
        return pd.read_sql(self.str_query, connection)

    def conn_databricks(self, bl_conn=False, int_max_wait=480, str_error='CONNECTION TIMEOUT EXPIRED',
                        bl_kill_process_when_databricks_down=False):
        """
        DOCSTRING: CONNECTION WITH DATABRICKS IN THE FIRST VALID DSN
        INPUTS: QUERT, LIST OF DSNS, LOGGER
        OUTPUTS: PANDAS DATAFRAME
        """
        # looping through dsns configuring
        for dsn in self.list_dsns:
            try:
                #   creating connection with the sql database
                connection = self.conn_dsn_databricks(dsn)
                df_databricks = func_timeout(int_max_wait, self.fetch_data_from_databricks,
                                             args=(connection))
                #   if the connection was established break the loop, otherwise log error
                bl_conn = True
                break
            except FunctionTimedOut:
                if self.logger != None:
                    CreateLog().warning(self.logger, 'Connection could not be established in '
                                         + 'the DSN {}'.format(dsn))
            except Exception as e:
                if self.logger != None:
                    CreateLog().warning(self.logger, 'Connection could not be established '
                                         + 'in the DSN {}. Error: {}. '.format(dsn, e))
        # caso a conexão não tenha sido estabelecida retornar um erro
        if bl_conn == False:
            if self.logger != None:
                CreateLog().warning(self.logger, 'Connection to Databricks could not be established '
                                     + 'in any of the DSNs, please validate the stability of the service. List of DSNs '
                                     + 'configured on the machine: {}'.format(self.list_dsns))
                if bl_kill_process_when_databricks_down == True:
                    raise Exception(self.logger, 'Connection to Databricks could not be established '
                                    + 'in any of the DSNs, please validate the stability of the service. List of DSNs '
                                    + 'configured on the machine: {}'.format(self.list_dsns))
            return str_error
        # return desired dataframe
        return df_databricks

    def retrieve_query_data(self, int_max_wait=480, i=0, df_databricks=''):
        """
        DOCSTRING: RETRIEVE QUERY DATA
        INPUTS: -
        OUTPUTS: DATAFRAME
        """
        # connection class
        class_databricks = Databricks(
            self.str_query, self.list_dsns, self.logger)
        # looping until overclock iterator
        while (type(df_databricks) == str) and (i <= self.max_error_attempts):
            #   attempting connection
            df_databricks = class_databricks.conn_databricks(
                int_max_wait=int_max_wait)
            #   logger-wise, if it is None print message, otherwise call the logger object to store
            #       data in network
            if self.logger is None:
                print('#{} Attempting connection with DSNs to Databricks'.format(i))
            else:
                CreateLog().info(self.logger,
                                  '#{} Attempting connection with DSNs to Databricks'.format(i))
            #   looping iterator to next count
            i += 1
        # return dataframe or error message
        return df_databricks



import os
from io import BytesIO, RawIOBase, BufferedIOBase
from typing import BinaryIO, Optional, Dict, Any, Union, IO, AnyStr
from minio import Minio
from minio.error import S3Error
from logging import Logger
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class MinioClient:

    def __init__(
        self,
        user: str,
        password: str,
        endpoint: Optional[str] = "localhost:9000",
        bl_secure: Optional[bool] = True,
        logger: Optional[Logger] = None
    ):
        self.client = Minio(
            endpoint,
            access_key=user,
            secret_key=password,
            secure=bl_secure
        )
        self.logger = logger

    def _log_info(self, message: str):
        if self.logger is not None:
            CreateLog().info(self.logger, message)
        else:
            print(f"INFO: {message}")

    def _log_critical(self, message: str):
        if self.logger is not None:
            CreateLog().critical(self.logger, message)
        else:
            print(f"CRITICAL: {message}")

    def make_bucket(self, bucket_name: str) -> bool:
        try:
            if not self.client.bucket_exists(bucket_name):
                self.client.make_bucket(bucket_name)
                self._log_info(f"Bucket {bucket_name} created successfully.")
            return True
        except S3Error as e:
            self._log_critical(f"Error creating bucket {bucket_name}: {e}")
            return False

    def put_object_from_file(
        self,
        bucket_name: str,
        object_name: str,
        file_path: str,
        dict_metadata: Optional[Dict[str, Any]] = None,
        content_type: Optional[str] = None
    ) -> bool:
        if not os.path.exists(file_path):
            self._log_critical(f"File {file_path} does not exist.")
            return False
        if not self.make_bucket(bucket_name):
            return False
        try:
            self.client.fput_object(
                bucket_name,
                object_name,
                file_path,
                metadata=dict_metadata,
                content_type=content_type
            )
            self._log_info(
                f"File '{file_path}' uploaded as '{object_name}' to bucket '{bucket_name}'.")
            return True
        except S3Error as e:
            self._log_critical(
                f"Error uploading file '{file_path}' as '{object_name}' to bucket '{bucket_name}': {e}")
            return False

    def put_object_from_stream(
        self,
        bucket_name: str,
        object_name: str,
        stream: Union[IO[AnyStr], BinaryIO, RawIOBase, BufferedIOBase],
        int_lenght: int,
        dict_metadata: Optional[Dict[str, Any]] = None,
        content_type: Optional[str] = None,
        seek_to_start: bool = None
    ) -> bool:
        if not self.make_bucket(bucket_name):
            return False
        try:
            if seek_to_start and hasattr(stream, 'seekable') and stream.seekable():
                stream.seek(0)
            self.client.put_object(bucket_name, object_name, stream, int_lenght,
                                metadata=dict_metadata, content_type=content_type)
            self._log_info(f"Stream uploaded as '{object_name}' to bucket '{bucket_name}'")
            return True
        except (S3Error, OSError) as e:
            self._log_critical(f"Error uploading stream as '{object_name}': {e}")
            return False

    def put_object_from_bytes(
        self,
        bucket_name: str,
        object_name: str,
        data: bytes,
        dict_metadata: Optional[Dict[str, Any]] = None,
        content_type: Optional[str] = None
    ) -> bool:
        if not self.make_bucket(bucket_name):
            return False
        try:
            with BytesIO(data) as stream:
                self.client.put_object(bucket_name, object_name, stream, len(data),
                                       metadata=dict_metadata, content_type=content_type)
                self._log_info(f"Bytes data uploaded as '{object_name}' to bucket '{bucket_name}'")
                return True
        except S3Error as e:
            self._log_critical(f"Error uploading bytes as '{object_name}': {e}")
            return False

    def get_object_as_bytes(
        self,
        bucket_name: str,
        object_name: str,
    ) -> Optional[bytes]:
        try:
            response = self.client.get_object(bucket_name, object_name)
            data = response.read()
            self._log_info(f"Successfully retrieved object '{object_name}' from bucket '{bucket_name}'")
            return data
        except S3Error as e:
            self._log_critical(f"Error retrieving object '{object_name}' from bucket '{bucket_name}': {e}")
            return None
        finally:
            if 'response' in locals():
                response.close()
                response.release_conn()

    def get_object_to_file(
        self,
        bucket_name: str,
        object_name: str,
        file_path: str,
    ) -> bool:
        try:
            self.client.fget_object(bucket_name, object_name, file_path)
            self._log_info(f"Successfully downloaded object '{object_name}' to '{file_path}'")
            return True
        except S3Error as e:
            self._log_critical(f"Error downloading object '{object_name}': {e}")
            return False

    def list_objects(
        self,
        bucket_name: str,
        bl_include_version: bool = False,
        prefix: Optional[str] = None,
        recursive: bool = False
    ) -> Optional[list]:
        try:
            objects = self.client.list_objects(bucket_name, include_version=bl_include_version,
                                               prefix=prefix, recursive=recursive)
            return [obj.object_name for obj in objects]
        except S3Error as e:
            self._log_critical(f"Error listing objects in bucket '{bucket_name}': {e}")
            return None



from typing import List, Any
from stpstone.dsa.queues.simple_queue import Queue


class JosephusSolver:
    """
    A class to solve the Josephus problem, which simulates an elimination process
    where every k-th participant is removed until only one survivor remains.
    """

    def __init__(self, list_: List[Any], inst_steps: int) -> None:
        self.list_ = list_
        self.inst_steps = inst_steps

    @property
    def last_survivor(self) -> Any:
        cls_queue = Queue()
        for item in self.list_:
            cls_queue.enqueue(item)
        while cls_queue.size > 1:
            for _ in range(self.inst_steps):
                cls_queue.enqueue(cls_queue.dequeue)
            cls_queue.dequeue
        return cls_queue.dequeue



import psycopg2
import subprocess
import os
import pandas as pd
from psycopg2.extensions import connection as Connection, cursor as Cursor
from psycopg2.extras import execute_values
from logging import Logger
from typing import List, Dict, Any, Optional
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.parsers.json import JsonFiles
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.connections.databases.abc import DatabaseConnection


class PostgreSQLDB(DatabaseConnection):

    def __init__(self, dbname:str, user:str, password:str, host:str, port:int,
                 str_schema:str="public", logger:Optional[Logger]=None) -> None:
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.str_schema = str_schema
        self.logger = logger
        self.dict_db_config = {
            "dbname": self.dbname,
            "user": self.user,
            "password": self.password,
            "host": self.host,
            "port": self.port
        }
        self.conn: Connection = psycopg2.connect(**self.dict_db_config)
        self.cursor: Cursor = self.conn.cursor()
        self.execute(f"SET search_path TO '{self.str_schema}';")

    def execute(self, str_query:str) -> None:
        self.cursor.execute(str_query)

    def read(self, str_query:str, dict_type_cols:Optional[Dict[str, Any]]=None,
              list_cols_dt:Optional[List[str]]=None, str_fmt_dt:Optional[str]=None) -> pd.DataFrame:
        # retrieving dataframe
        df_ = pd.read_sql_query(str_query, self.conn)
        # changing data types
        if all([x is not None for x in [dict_type_cols, list_cols_dt, str_fmt_dt]]):
            df_ = df_.astype(dict_type_cols)
            for col_ in list_cols_dt:
                df_[col_] = [DatesBR().str_date_to_datetime(d, str_fmt_dt) for d in df_[col_]]
        # return dataframe
        return df_

    def insert(self, json_data:List[Dict[str, Any]], str_table_name:str,
        bl_insert_or_ignore:bool=False) -> None:
        # validate json, in order to have the same keys
        json_data = JsonFiles().normalize_json_keys(json_data)
        # sql insert statement
        list_columns = ", ".join(json_data[0].keys())
        list_placeholders = ", ".join(["%s" for _ in json_data[0]])
        if bl_insert_or_ignore == True:
            str_query = f"INSERT INTO {str_table_name} ({list_columns}) VALUES " \
                + f"({list_placeholders}) ON CONFLICT DO NOTHING"
        else:
            str_query = f"INSERT INTO {str_table_name} ({list_columns}) VALUES " \
                + f"({list_placeholders})"
        try:
            execute_values(
                self.cursor,
                str_query,
                [tuple(record.values()) for record in json_data]
            )
            self.conn.commit()
            if self.logger is not None:
                CreateLog().info(
                    self.logger,
                    f"Successful commit in db {self.dict_db_config["dbname"]} "
                    + f"/ table {str_table_name}."
                )
        except Exception as e:
            self.conn.rollback()
            self.close
            if self.logger is not None:
                CreateLog().error(
                    self.logger,
                    "ERROR WHILE INSERTING DATA\n"
                    + f"DB_CONFIG: {self.dict_db_config}\n"
                    + f"TABLE_NAME: {str_table_name}\n"
                    + f"JSON_DATA: {json_data}\n"
                    + f"ERROR_MESSAGE: {e}"
                )
            raise Exception(
                "ERROR WHILE INSERTING DATA\n"
                + f"DB_CONFIG: {self.dict_db_config}\n"
                + f"TABLE_NAME: {str_table_name}\n"
                + f"JSON_DATA: {json_data}\n"
                + f"ERROR_MESSAGE: {e}"
            )

    def close(self) -> None:
        self.conn.close()

    def backup(self, str_backup_dir:str, str_bkp_name:str=None) -> str:
        try:
            # ensure the backup directory exists
            os.makedirs(str_backup_dir, exist_ok=True)
            # generate the backup file name
            backup_file = os.path.join(str_backup_dir, str_bkp_name)
            env = os.environ.copy()
            env["PGPASSWORD"] = self.password
            # run the pg_dump command
            command = [
                "pg_dump",
                "-h", self.host,
                "-p", str(self.port),
                "-U", self.user,
                "-F", "c",  # custom format
                "-b",       # include large objects
                "-f", backup_file,  # output file
                self.dbname
            ]
            subprocess.run(command, check=True, env=env)
            return f"Backup successful! File saved at: {backup_file}"
        except subprocess.CalledProcessError as e:
            return f"Backup failed: {e}"
        except Exception as e:
            return f"An error occurred: {e}"


### WEBHOOK SLACK ###

from requests import request
from stpstone.utils.parsers.json import JsonFiles


class WebhookSlack:

    def __init__(self, url_webhook:str, id_channel:str, str_username:str='webhookbot',
                 str_icon_emoji:str=':bricks:') -> None:
        self.url_webhook = url_webhook
        self.id_channel = id_channel
        self.str_username = str_username
        self.str_icon_emoji = str_icon_emoji

    def send_message(self, str_msg:str, str_method:str='POST') -> None:
        """
        DOCSTRING:
        INPUTS:
            - MESSAGE:STR
            - METHOD:STR
            - ICON_EMOJI:STR (AVAILABLE EMOJIS: https://raw.githubusercontent.com/iamcal/emoji-data/master/emoji.json)
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        dict_payload = {
            'channel': self.id_channel,
            'username': self.str_username,
            'text': str_msg,
            'icon_emoji': self.str_icon_emoji
        }
        dict_headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        resp_req = request(str_method, self.url_webhook, headers=dict_headers,
                           data=JsonFiles().dict_to_json(dict_payload))
        resp_req.raise_for_status()
        return resp_req.text



### PARALLEL PIPELINES ###

from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Any


def parallelpipeline(data:Any, functions:List[Callable]) -> Any:
    """
    Executes a sequence of functions in parallel where possible
    Examples of usage:
        Batch processing, image transformation
    Args:
        data (Any): Initial data input
        functions (List[Callable]): A list of functions to apply
    Returns:
        Any: The processed data
    """
    with ThreadPoolExecutor() as executor:
        results = list(executor.map(lambda f: f(data), functions))
    return results[-1]  # Return last transformation


if __name__ == "__main__":
    def add_one(x): return x + 1
    def multiply_by_two(x): return x * 2
    def subtract_five(x): return x - 5

    result = parallelpipeline(5, [add_one, multiply_by_two, subtract_five])
    print(result)
